from __future__ import annotations

from enum import Enum
import re
from typing import Any

from isic_metadata.diagnosis_hierarchical import DiagnosisEnum  # noqa: F401


class ClinSizeLongDiamMm:
    @classmethod
    def parse_measurement_str(cls, value: Any) -> Any:
        if isinstance(value, str):
            match = re.match(r"(.+)(um|mm|cm)$", value)

            if not match:
                return value

            float_value, units = match.groups()
            float_value = float(float_value)

            # Convert to mm
            if units == "um":
                float_value *= 1e-3
            elif units == "cm":
                float_value *= 1e1

            return round(float_value, ndigits=1)

        return value


class Age:
    @classmethod
    def handle_85plus(cls, value: Any) -> Any:
        if value == "85+":
            return 85

        return value

    @staticmethod
    def clamp_upper_bound(value: int) -> int:
        return min(value, 85)


class DiagnosisConfirmTypeEnum(str, Enum):
    histopathology = "histopathology"
    serial_imaging_showing_no_change = "serial imaging showing no change"
    single_image_expert_consensus = "single image expert consensus"
    confocal_microscopy_with_consensus_dermoscopy = "confocal microscopy with consensus dermoscopy"
    single_contributor_clinical_assessment = "single contributor clinical assessment"


class LegacyDxEnum(str, Enum):
    actinic_keratosis = "actinic keratosis"
    adnexal_tumor = "adnexal tumor"
    aimp = "AIMP"
    angiokeratoma = "angiokeratoma"
    angioma = "angioma"
    basal_cell_carcinoma = "basal cell carcinoma"
    cafe_au_lait_macule = "cafe-au-lait macule"
    dermatofibroma = "dermatofibroma"
    ephelis = "ephelis"
    lentigo_nos = "lentigo NOS"
    lentigo_simplex = "lentigo simplex"
    lichenoid_keratosis = "lichenoid keratosis"
    melanoma = "melanoma"
    melanoma_metastasis = "melanoma metastasis"
    merkel_cell_carcinoma = "merkel cell carcinoma"
    mucosal_melanosis = "mucosal melanosis"
    nevus = "nevus"
    nevus_spilus = "nevus spilus"
    seborrheic_keratosis = "seborrheic keratosis"
    solar_lentigo = "solar lentigo"
    squamous_cell_carcinoma = "squamous cell carcinoma"
    clear_cell_acanthoma = "clear cell acanthoma"
    atypical_spitz_tumor = "atypical spitz tumor"
    acrochordon = "acrochordon"
    angiofibroma_or_fibrous_papule = "angiofibroma or fibrous papule"
    neurofibroma = "neurofibroma"
    pyogenic_granuloma = "pyogenic granuloma"
    scar = "scar"
    sebaceous_adenoma = "sebaceous adenoma"
    sebaceous_hyperplasia = "sebaceous hyperplasia"
    verruca = "verruca"
    atypical_melanocytic_proliferation = "atypical melanocytic proliferation"
    epidermal_nevus = "epidermal nevus"
    pigmented_benign_keratosis = "pigmented benign keratosis"
    vascular_lesion = "vascular lesion"
    other = "other"


class ImageTypeEnum(str, Enum):
    dermoscopic = "dermoscopic"
    clinical_overview = "clinical: overview"
    clinical_close_up = "clinical: close-up"
    tbp_tile_close_up = "TBP tile: close-up"
    tbp_tile_overview = "TBP tile: overview"
    # Note that RCM types need to be added to validate_rcm_fields
    rcm_macroscopic = "RCM: macroscopic"
    rcm_tile = "RCM: tile"
    rcm_mosaic = "RCM: mosaic"


class DermoscopicTypeEnum(str, Enum):
    contact_polarized = "contact polarized"
    contact_non_polarized = "contact non-polarized"
    non_contact_polarized = "non-contact polarized"


class TBPTileTypeEnum(str, Enum):
    tbp_3d_white = "3D: white"
    tbp_3d_xp = "3D: XP"
    tbp_2d = "2D"


class MelThickMm:
    _regex = re.compile(r"^([\d.]+)(\s+)?(mm)?$")

    @classmethod
    def parse_measurement_str(cls, value: Any) -> Any:
        if isinstance(value, str):
            result = re.match(MelThickMm._regex, value)

            if not result:
                return value

            value = result.group(1)

            return float(value)

        return value


class MelMitoticIndexEnum(str, Enum):
    zero = "0/mm^2"
    lt_one = "<1/mm^2"
    one = "1/mm^2"
    two = "2/mm^2"
    three = "3/mm^2"
    four = "4/mm^2"
    gt_4 = ">4/mm^2"


class AnatomSiteGeneralEnum(str, Enum):
    head_neck = "head/neck"
    upper_extremity = "upper extremity"
    lower_extremity = "lower extremity"
    anterior_torso = "anterior torso"
    posterior_torso = "posterior torso"
    palms_soles = "palms/soles"
    lateral_torso = "lateral torso"
    oral_genital = "oral/genital"


class AnatomSiteSpecialEnum(str, Enum):
    acral_nos = "acral NOS"
    nail_nos = "nail NOS"
    fingernail = "fingernail"
    toenail = "toenail"
    acral_palms_soles = "acral palms or soles"
    oral_genital = "oral or genital"


class ColorTintEnum(str, Enum):
    blue = "blue"
    pink = "pink"
    none = "none"


class FitzpatrickSkinType(str, Enum):
    type_i = "I"
    type_ii = "II"
    type_iii = "III"
    type_iv = "IV"
    type_v = "V"
    type_vi = "VI"


class ImageManipulationEnum(str, Enum):
    instrument_only = "instrument only"
    altered = "altered"
    synthetic = "synthetic"



from pydantic import ValidationError
import pytest

from isic_metadata.diagnosis_hierarchical import DiagnosisEnum
from isic_metadata.metadata import MetadataRow


@pytest.mark.parametrize(
    ("raw", "parsed"),
    [
        ("Benign", ["Benign"]),
        ("Benign - Other", ["Benign", "Benign - Other"]),
        ("Blue nevus", ["Benign", "Benign melanocytic proliferations", "Nevus", "Blue nevus"]),
        (
            "Squamous cell carcinoma, NOS",
            ["Malignant", "Malignant epidermal proliferations", "Squamous cell carcinoma, NOS"],
        ),
        (
            "Blue nevus, Sclerosing",
            [
                "Benign",
                "Benign melanocytic proliferations",
                "Nevus",
                "Blue nevus",
                "Blue nevus, Sclerosing",
            ],
        ),
    ],
)
def test_diagnosis(raw, parsed):
    metadata = MetadataRow.model_validate({"diagnosis": raw})

    for i, diagnosis in enumerate(parsed, start=1):
        assert getattr(metadata, f"diagnosis_{i}") == diagnosis


def test_top_level_diagnosis_is_never_exported():
    metadata = MetadataRow.model_validate({"diagnosis": "Benign"})
    assert "diagnosis" not in metadata.model_dump()
    assert metadata.diagnosis_1 == "Benign"


def test_diagnosis_enum_has_unique_terminal_values():
    terminal_nodes = [member.value.split(":")[-1] for member in DiagnosisEnum]
    assert len(terminal_nodes) == len(set(terminal_nodes))


def test_single_value_diagnosis_is_favored():
    # test that passing in a single diagnosis value is favored over multiple values. used
    # for when data is coming from the database and potentially contains an existing
    # 1..5 diagnosis and a newly updated single diagnosis.
    with pytest.raises(ValidationError) as excinfo:
        MetadataRow.model_validate(
            {
                "diagnosis": "Fibroma, Sclerotic",
                "mel_ulcer": True,
                # these should be ignored
                "diagnosis_1": "Benign",
                "diagnosis_2": "Benign melanocytic proliferations",
                "diagnosis_3": "Nevus",
            }
        )
    assert "Setting mel_ulcer is incompatible with diagnosis" in excinfo.value.errors()[0]["msg"]


def test_diagnosis_multiple_levels_is_coerced():
    # test that passing in diagnosis_1..5 is coerced into a single diagnosis field to handle
    # cross field input validation
    metadata = MetadataRow.model_validate({"diagnosis_1": "Benign"})
    assert metadata.diagnosis_1 == "Benign"
    assert metadata.diagnosis_2 is None
    assert metadata.diagnosis_3 is None
    assert metadata.diagnosis_4 is None
    assert metadata.diagnosis_5 is None


def test_diagnosis_validation_is_idempotent():
    # test that running model_validate on a MetadataRow multiple times does not change the
    # output
    metadata = MetadataRow.model_validate({"diagnosis": "Melanoma Invasive"})
    assert metadata.diagnosis_1 == "Malignant"
    assert metadata.diagnosis_2 == "Malignant melanocytic proliferations (Melanoma)"
    assert metadata.diagnosis_3 == "Melanoma Invasive"
    metadata_2 = MetadataRow.model_validate(
        metadata.model_dump(exclude_unset=True, exclude_none=True, exclude={"unstructured"})
    )
    assert metadata == metadata_2



from __future__ import annotations

from collections import defaultdict
from decimal import Decimal
from typing import Annotated, Any, Literal

from annotated_types import Ge
from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    ValidationError,
    computed_field,
    field_validator,
    model_validator,
)
from pydantic_core import ErrorDetails, PydanticCustomError

from isic_metadata.fields import (
    Age,
    AnatomSiteGeneralEnum,
    AnatomSiteSpecialEnum,
    ClinSizeLongDiamMm,
    ColorTintEnum,
    DermoscopicTypeEnum,
    DiagnosisConfirmTypeEnum,
    DiagnosisEnum,
    FitzpatrickSkinType,
    ImageManipulationEnum,
    ImageTypeEnum,
    MelMitoticIndexEnum,
    MelThickMm,
    TBPTileTypeEnum,
)

"""
IGNORE_RCM_MODEL_CHECKS is a special attribute that allows disabling RCM model checks. It's
provided to allow batch checks to be performed on a set of metadata rows, some of which may
not be valid on a row level. e.g. if a batch of metadata rows contains RCM case ids that are
inconsistent with lesion ids, that ought to be validated even if the row with an RCM case id
has a missing image_type for instance.
"""
IGNORE_RCM_MODEL_CHECKS = "_ignore_rcm_model_checks"

CUSTOM_MESSAGES = {
    "enum": "Unsupported value for {loc}: '{value}'.",
    "int_parsing": "Unable to parse value as an integer.",
    "bool_parsing": "Unable to parse value as a boolean.",
    "decimal_parsing": "Unable to parse value as a number.",
    "greater_than_equal": "Number must be greater than or equal to {ge}.",
}


def convert_errors(e: ValidationError) -> list[ErrorDetails]:
    new_errors: list[ErrorDetails] = []
    for error in e.errors():
        custom_message = CUSTOM_MESSAGES.get(error["type"])
        if custom_message:
            ctx = error.get("ctx")
            loc = {"loc": error.get("loc")[0]}
            if ctx:
                ctx.update(loc)
                ctx.update({"value": error["input"]})

            error["msg"] = custom_message.format(**ctx) if ctx else custom_message
        new_errors.append(error)
    return new_errors


def error_incompatible_fields(
    field1: str, field2: str, field1_value: str | None = None, field2_value: str | None = None
) -> PydanticCustomError:
    message = (
        "Setting {{field1}}{maybe_field1_value} is incompatible with {{field2}}"
        "{maybe_field2_value}."
    ).format(
        maybe_field1_value=" to {field1_value}" if field1_value else "",
        maybe_field2_value=" {field2_value}" if field2_value else "",
    )

    return PydanticCustomError(
        "incompatible_fields",
        message,
        {
            "field1": field1,
            "field1_value": field1_value,
            "field2": field2,
            "field2_value": field2_value,
        },
    )


def error_missing_field(
    field1: str, field2: str, field1_value: str | None = None, field2_value: str | None = None
) -> PydanticCustomError:
    message = (
        "Setting {{field1}}{maybe_field1_value} requires setting {{field2}}{maybe_field2_value}."
    ).format(
        maybe_field1_value=" to {field1_value}" if field1_value else "",
        maybe_field2_value=" to {field2_value}" if field2_value else "",
    )

    return PydanticCustomError(
        "missing_field",
        message,
        {
            "field1": field1,
            "field1_value": field1_value,
            "field2": field2,
            "field2_value": field2_value,
        },
    )


class MetadataBatch(BaseModel):
    """
    A batch of metadata rows.

    This is useful for performing checks that span across multiple rows.
    """

    items: list[MetadataRow]

    @model_validator(mode="after")
    def check_patients_lesions(self) -> MetadataBatch:
        lesion_to_patients: dict[str, set[str]] = defaultdict(set)

        for item in self.items:
            if item.patient_id and item.lesion_id:
                lesion_to_patients[item.lesion_id].add(item.patient_id)

        bad_lesions = [
            lesion for lesion in lesion_to_patients if len(lesion_to_patients[lesion]) > 1
        ]
        if bad_lesions:
            raise PydanticCustomError(
                "one_lesion_multiple_patients",
                "One or more lesions belong to multiple patients.",
                {"examples": bad_lesions[:5]},
            )

        return self

    @model_validator(mode="after")
    def check_rcm_at_most_one_macroscopic_image(self) -> MetadataBatch:
        rcm_case_to_macroscopic: dict[str, int] = defaultdict(int)

        for item in self.items:
            if item.rcm_case_id and item.image_type == ImageTypeEnum.rcm_macroscopic:
                rcm_case_to_macroscopic[item.rcm_case_id] += 1

        bad_rcm_cases = [
            rcm_case
            for rcm_case in rcm_case_to_macroscopic
            if rcm_case_to_macroscopic[rcm_case] > 1
        ]
        if bad_rcm_cases:
            raise PydanticCustomError(
                "rcm_multiple_macroscopics",
                "One or more RCM cases have multiple macroscopic images.",
                {"examples": bad_rcm_cases[:5]},
            )

        return self

    @model_validator(mode="after")
    def check_rcm_case_at_most_one_lesion(self) -> MetadataBatch:
        rcm_case_to_lesions: dict[str, set[str]] = defaultdict(set)

        for item in self.items:
            if item.rcm_case_id and item.lesion_id:
                rcm_case_to_lesions[item.rcm_case_id].add(item.lesion_id)

        bad_rcm_cases = [
            rcm_case for rcm_case in rcm_case_to_lesions if len(rcm_case_to_lesions[rcm_case]) > 1
        ]
        if bad_rcm_cases:
            raise PydanticCustomError(
                "one_rcm_case_multiple_lesions",
                "One or more RCM cases belong to multiple lesions.",
                {"examples": bad_rcm_cases[:5]},
            )

        return self


class MetadataRow(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        # This is useful for patient/lesion id fields, which could be numeric
        coerce_numbers_to_str=True,
        # extra values should be gathered by build_extra, so there's no need to allow them here
        extra="forbid",
    )

    age: (
        Annotated[
            int, BeforeValidator(Age.handle_85plus), AfterValidator(Age.clamp_upper_bound), Ge(0)
        ]
        | None
    ) = None
    sex: Literal["male", "female"] | None = None
    anatom_site_general: AnatomSiteGeneralEnum | None = None
    anatom_site_special: AnatomSiteSpecialEnum | None = None
    diagnosis: (
        Annotated[
            DiagnosisEnum,
            BeforeValidator(DiagnosisEnum.accept_terminal_values),
        ]
        | None
    ) = Field(default=None, exclude=True)  # never expose the fully qualified diagnosis
    diagnosis_confirm_type: DiagnosisConfirmTypeEnum | None = None
    personal_hx_mm: bool | None = None
    family_hx_mm: bool | None = None
    clin_size_long_diam_mm: (
        Annotated[Decimal, BeforeValidator(ClinSizeLongDiamMm.parse_measurement_str)] | None
    ) = None
    fitzpatrick_skin_type: FitzpatrickSkinType | None = None
    melanocytic: bool | None = None
    concomitant_biopsy: bool | None = None

    mel_mitotic_index: MelMitoticIndexEnum | None = None
    mel_thick_mm: Annotated[Decimal, BeforeValidator(MelThickMm.parse_measurement_str)] | None = (
        None
    )
    mel_ulcer: bool | None = None

    patient_id: str | None = None
    lesion_id: str | None = None
    acquisition_day: int | None = None

    image_manipulation: ImageManipulationEnum | None = None

    image_type: ImageTypeEnum | None = None
    dermoscopic_type: DermoscopicTypeEnum | None = None
    tbp_tile_type: TBPTileTypeEnum | None = None

    rcm_case_id: str | None = None

    unstructured: dict[str, Any] = {}

    # Unused and undocumented
    marker_pen: bool | None = None
    hairy: bool | None = None
    blurry: bool | None = None
    color_tint: ColorTintEnum | None = None

    @computed_field
    def diagnosis_1(self) -> str | None:
        return DiagnosisEnum.levels(self.diagnosis)[0] if self.diagnosis else None

    @computed_field
    def diagnosis_2(self) -> str | None:
        return DiagnosisEnum.levels(self.diagnosis)[1] if self.diagnosis else None

    @computed_field
    def diagnosis_3(self) -> str | None:
        return DiagnosisEnum.levels(self.diagnosis)[2] if self.diagnosis else None

    @computed_field
    def diagnosis_4(self) -> str | None:
        return DiagnosisEnum.levels(self.diagnosis)[3] if self.diagnosis else None

    @computed_field
    def diagnosis_5(self) -> str | None:
        return DiagnosisEnum.levels(self.diagnosis)[4] if self.diagnosis else None

    __slots__ = (IGNORE_RCM_MODEL_CHECKS,)

    # see https://github.com/pydantic/pydantic/issues/655#issuecomment-570312649 for details on
    # implementing a private property to be used internally.
    def __init__(self, **kwargs) -> None:
        object.__setattr__(self, IGNORE_RCM_MODEL_CHECKS, False)

        if IGNORE_RCM_MODEL_CHECKS in kwargs:
            object.__setattr__(self, IGNORE_RCM_MODEL_CHECKS, kwargs.pop(IGNORE_RCM_MODEL_CHECKS))

        super().__init__(**kwargs)

    @model_validator(mode="before")
    @classmethod
    def handle_hierarchical_diagnosis_modes_and_unstructured_fields(
        cls, values: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Handle the case where hierarchical diagnosis values are passed in as multiple fields.

        Practically, ingesting data should never pass in multiple values but instead use the
        colon-separated `diagnosis` field. This method is provided for the scenario where
        data needs to be retrieved from the database (where it's stored multi-valued) and
        revalidated. This method also handles putting any unrecognized fields into an unstructured
        field. Unfortunately, pydantic doesn't yet support ordering different model validators so
        these both need to be combined into one method.
        """
        using_diagnoses_multi_values = any(f"diagnosis_{i}" in values for i in range(1, 6))
        using_diagnosis_single_value = bool(values.get("diagnosis"))

        if using_diagnoses_multi_values and using_diagnosis_single_value:
            [values.pop(f"diagnosis_{i}", "") for i in range(1, 6)]
        elif using_diagnoses_multi_values:
            values["diagnosis"] = ":".join(values.pop(f"diagnosis_{i}", "") for i in range(1, 6))
            values["diagnosis"] = values["diagnosis"].rstrip(":")

        # handle unstructured fields
        # See https://github.com/samuelcolvin/pydantic/issues/2285 for more detail
        structured_field_names = {field for field in cls.model_fields if field != "unstructured"}

        unstructured: dict[str, Any] = {}
        for field_name in list(values):
            if field_name not in structured_field_names:
                unstructured[field_name] = values.pop(field_name)
        values["unstructured"] = unstructured
        return values

    @field_validator("*", mode="before")
    @classmethod
    def strip(cls, v: Any) -> Any:
        # str_strip_whitespace doesn't seem to work for Literal values as it does for str and enum
        if isinstance(v, str):
            v = v.strip()

        # drop empty strings as though they were never passed
        return None if v == "" else v

    @field_validator(
        "anatom_site_general",
        "clin_size_long_diam_mm",
        "diagnosis_confirm_type",
        "mel_mitotic_index",
        "mel_thick_mm",
        "sex",
        mode="before",
    )
    @classmethod
    def lower(cls, v: Any) -> Any:
        if isinstance(v, str):
            v = v.lower()
        return v

    @model_validator(mode="after")
    def validate_melanoma_fields(self) -> MetadataRow:
        melanoma_fields: list[str] = [
            "mel_mitotic_index",
            "mel_thick_mm",
            "mel_ulcer",
        ]

        for field in melanoma_fields:
            if not getattr(self, field):
                continue

            if not self.diagnosis:
                raise error_missing_field(field, "diagnosis", field2_value=self.diagnosis)

            if not DiagnosisEnum.is_melanoma(self.diagnosis):
                raise error_incompatible_fields(
                    field, "diagnosis", field2_value=self.diagnosis.value
                )

        return self

    @model_validator(mode="after")
    def validate_rcm_fields(self) -> MetadataRow:
        if getattr(self, IGNORE_RCM_MODEL_CHECKS):
            return self

        if not self.rcm_case_id:
            return self

        if not self.image_type:
            raise error_missing_field("rcm_case_id", "image_type")

        if self.image_type not in [
            ImageTypeEnum.rcm_macroscopic,
            ImageTypeEnum.rcm_tile,
            ImageTypeEnum.rcm_mosaic,
        ]:
            raise error_incompatible_fields(
                "rcm_case_id", "image_type", field2_value=ImageTypeEnum.rcm_macroscopic
            )

        return self

    @model_validator(mode="after")
    def validate_dermoscopic_fields(self) -> MetadataRow:
        if not self.dermoscopic_type:
            return self

        if not self.image_type:
            raise error_missing_field("dermoscopic_type", "image_type")

        if self.image_type != ImageTypeEnum.dermoscopic:
            raise error_incompatible_fields(
                "dermoscopic_type", "image_type", field2_value="dermoscopic"
            )

        return self

    @model_validator(mode="after")
    def validate_anatom_site_special(self) -> MetadataRow:
        if not self.anatom_site_special:
            return self

        if not self.anatom_site_general:
            raise error_missing_field("anatom_site_special", "anatom_site_general")

        valid_combinations = {
            AnatomSiteSpecialEnum.acral_nos: [
                AnatomSiteGeneralEnum.upper_extremity,
                AnatomSiteGeneralEnum.lower_extremity,
                AnatomSiteGeneralEnum.palms_soles,
            ],
            AnatomSiteSpecialEnum.nail_nos: [
                AnatomSiteGeneralEnum.upper_extremity,
                AnatomSiteGeneralEnum.lower_extremity,
                AnatomSiteGeneralEnum.palms_soles,
            ],
            AnatomSiteSpecialEnum.fingernail: [
                AnatomSiteGeneralEnum.upper_extremity,
                AnatomSiteGeneralEnum.palms_soles,
            ],
            AnatomSiteSpecialEnum.toenail: [
                AnatomSiteGeneralEnum.lower_extremity,
                AnatomSiteGeneralEnum.palms_soles,
            ],
            AnatomSiteSpecialEnum.acral_palms_soles: [
                AnatomSiteGeneralEnum.upper_extremity,
                AnatomSiteGeneralEnum.lower_extremity,
                AnatomSiteGeneralEnum.palms_soles,
            ],
            AnatomSiteSpecialEnum.oral_genital: [
                AnatomSiteGeneralEnum.head_neck,
                AnatomSiteGeneralEnum.oral_genital,
                AnatomSiteGeneralEnum.lower_extremity,
                AnatomSiteGeneralEnum.anterior_torso,
                AnatomSiteGeneralEnum.posterior_torso,
            ],
        }

        if self.anatom_site_special.value not in valid_combinations:
            return self

        if self.anatom_site_general.value not in valid_combinations[self.anatom_site_special.value]:
            raise error_incompatible_fields(
                "anatom_site_special",
                "anatom_site_general",
                self.anatom_site_special.value,
                self.anatom_site_general.value,
            )

        return self

    @model_validator(mode="after")
    def validate_tbp_tile_fields(self) -> MetadataRow:
        if not self.tbp_tile_type:
            return self

        if not self.image_type:
            raise error_missing_field("tbp_tile_type", "image_type")

        if self.image_type not in [
            ImageTypeEnum.tbp_tile_close_up,
            ImageTypeEnum.tbp_tile_overview,
        ]:
            raise error_incompatible_fields(
                "tbp_tile_type", "image_type", field2_value=self.image_type.value
            )

        return self

    @model_validator(mode="after")
    def validate_concomitant_biopsy(self) -> MetadataRow:
        if self.concomitant_biopsy and (
            not self.diagnosis_confirm_type or self.diagnosis_confirm_type != "histopathology"
        ):
            raise error_missing_field(
                "concomitant_biopsy", "diagnosis_confirm_type", field2_value="histopathology"
            )

        return self



from __future__ import annotations

from enum import Enum
from functools import cache
from typing import Any

# ruff: noqa: E501


class DiagnosisEnum(str, Enum):
    benign = "Benign"
    indeterminate = "Indeterminate"
    malignant = "Malignant"
    benign_benign_other = "Benign:Benign - Other"
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine = (
        "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine"
    )
    benign_benign_adnexal_epithelial_proliferations_follicular = (
        "Benign:Benign adnexal epithelial proliferations - Follicular"
    )
    benign_benign_adnexal_epithelial_proliferations_sebaceous = (
        "Benign:Benign adnexal epithelial proliferations - Sebaceous"
    )
    benign_benign_epidermal_proliferations = "Benign:Benign epidermal proliferations"
    benign_benign_melanocytic_proliferations = "Benign:Benign melanocytic proliferations"
    benign_benign_soft_tissue_proliferations_adipocytic = (
        "Benign:Benign soft tissue proliferations - Adipocytic"
    )
    benign_benign_soft_tissue_proliferations_cartilagenous_and_ossifying = (
        "Benign:Benign soft tissue proliferations - Cartilagenous and ossifying"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic"
    )
    benign_benign_soft_tissue_proliferations_muscle_tissue_or_myofibroblastic = (
        "Benign:Benign soft tissue proliferations - Muscle tissue or myofibroblastic"
    )
    benign_benign_soft_tissue_proliferations_myoepithelial = (
        "Benign:Benign soft tissue proliferations - Myoepithelial"
    )
    benign_benign_soft_tissue_proliferations_neural = (
        "Benign:Benign soft tissue proliferations - Neural"
    )
    benign_benign_soft_tissue_proliferations_vascular = (
        "Benign:Benign soft tissue proliferations - Vascular"
    )
    benign_collision_only_benign_proliferations = "Benign:Collision - Only benign proliferations"
    benign_cysts = "Benign:Cysts"
    benign_exogenous = "Benign:Exogenous"
    benign_flat_melanotic_pigmentations_not_melanocytic_nevus = (
        "Benign:Flat melanotic pigmentations - not melanocytic nevus"
    )
    benign_hemorrhagic_lesions = "Benign:Hemorrhagic lesions"
    benign_inflammatory_or_infectious_diseases = "Benign:Inflammatory or infectious diseases"
    benign_langerhans_cell_proliferations = "Benign:Langerhans cell proliferations"
    benign_mast_cell_proliferations = "Benign:Mast cell proliferations"
    indeterminate_indeterminate_epidermal_proliferations = (
        "Indeterminate:Indeterminate epidermal proliferations"
    )
    indeterminate_indeterminate_melanocytic_proliferations = (
        "Indeterminate:Indeterminate melanocytic proliferations"
    )
    malignant_collision_at_least_one_malignant_proliferation = (
        "Malignant:Collision - At least one malignant proliferation"
    )
    malignant_lymphocytic_proliferations_b_cell = "Malignant:Lymphocytic proliferations - B-Cell"
    malignant_lymphocytic_proliferations_t_cellnk = (
        "Malignant:Lymphocytic proliferations - T-Cell/NK"
    )
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine = (
        "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine"
    )
    malignant_malignant_adnexal_epithelial_proliferations_follicular = (
        "Malignant:Malignant adnexal epithelial proliferations - Follicular"
    )
    malignant_malignant_adnexal_epithelial_proliferations_sebaceous = (
        "Malignant:Malignant adnexal epithelial proliferations - Sebaceous"
    )
    malignant_malignant_epidermal_proliferations = "Malignant:Malignant epidermal proliferations"
    malignant_malignant_melanocytic_proliferations_melanoma = (
        "Malignant:Malignant melanocytic proliferations (Melanoma)"
    )
    malignant_malignant_soft_tissue_proliferations_adipocytic = (
        "Malignant:Malignant soft tissue proliferations - Adipocytic"
    )
    malignant_malignant_soft_tissue_proliferations_cartilagenous_and_ossifying = (
        "Malignant:Malignant soft tissue proliferations - Cartilagenous and ossifying"
    )
    malignant_malignant_soft_tissue_proliferations_fibro_histiocytic = (
        "Malignant:Malignant soft tissue proliferations - Fibro-histiocytic"
    )
    malignant_malignant_soft_tissue_proliferations_muscle_tissue_or_myofibroblastic = (
        "Malignant:Malignant soft tissue proliferations - Muscle tissue or myofibroblastic"
    )
    malignant_malignant_soft_tissue_proliferations_myoepithelial = (
        "Malignant:Malignant soft tissue proliferations - Myoepithelial"
    )
    malignant_malignant_soft_tissue_proliferations_neural = (
        "Malignant:Malignant soft tissue proliferations - Neural"
    )
    malignant_malignant_soft_tissue_proliferations_unknown_or_other_histiogenesis = (
        "Malignant:Malignant soft tissue proliferations - Unknown or other histiogenesis"
    )
    malignant_malignant_soft_tissue_proliferations_vascular = (
        "Malignant:Malignant soft tissue proliferations - Vascular"
    )
    malignant_merkel_cell_proliferation = "Malignant:Merkel cell proliferation"
    malignant_skin_metastasis_of_internal_solid_cancer_non_hematological = (
        "Malignant:Skin metastasis of internal solid cancer - non-hematological"
    )
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_apocrine_tubular_adenoma = "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Apocrine tubular adenoma"
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_cylindoma = (
        "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Cylindoma"
    )
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_cystadenoma = (
        "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Cystadenoma"
    )
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_fibroadenoma = (
        "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Fibroadenoma"
    )
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_hidradenoma_papilliferum = "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Hidradenoma papilliferum"
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_hidradenoma_apocrine = "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Hidradenoma, Apocrine"
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_hidradenoma = (
        "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Hidradenoma"
    )
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_mixed_tumor = (
        "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Mixed tumor"
    )
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_poroma = (
        "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Poroma"
    )
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_spiradenoma = (
        "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Spiradenoma"
    )
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_supernumerary_nipple = (
        "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Supernumerary nipple"
    )
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_syringocystadenoma_papilliferum = "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Syringocystadenoma papilliferum"
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_syringofibroadenoma = (
        "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Syringofibroadenoma"
    )
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_syringoma = (
        "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Syringoma"
    )
    benign_benign_adnexal_epithelial_proliferations_follicular_folliculosebaceous_cystic_hamartoma = "Benign:Benign adnexal epithelial proliferations - Follicular:Folliculosebaceous cystic hamartoma"
    benign_benign_adnexal_epithelial_proliferations_follicular_nevus_comedonicus = (
        "Benign:Benign adnexal epithelial proliferations - Follicular:Nevus comedonicus"
    )
    benign_benign_adnexal_epithelial_proliferations_follicular_panfolliculoma = (
        "Benign:Benign adnexal epithelial proliferations - Follicular:Panfolliculoma"
    )
    benign_benign_adnexal_epithelial_proliferations_follicular_pilar_sheath_acanthoma = (
        "Benign:Benign adnexal epithelial proliferations - Follicular:Pilar sheath acanthoma"
    )
    benign_benign_adnexal_epithelial_proliferations_follicular_pilomatricoma = (
        "Benign:Benign adnexal epithelial proliferations - Follicular:Pilomatricoma"
    )
    benign_benign_adnexal_epithelial_proliferations_follicular_proliferating_tricholemmal_tumor = "Benign:Benign adnexal epithelial proliferations - Follicular:Proliferating tricholemmal tumor"
    benign_benign_adnexal_epithelial_proliferations_follicular_trichoblastoma = (
        "Benign:Benign adnexal epithelial proliferations - Follicular:Trichoblastoma"
    )
    benign_benign_adnexal_epithelial_proliferations_follicular_trichoepithelioma = (
        "Benign:Benign adnexal epithelial proliferations - Follicular:Trichoepithelioma"
    )
    benign_benign_adnexal_epithelial_proliferations_follicular_trichofolliculoma = (
        "Benign:Benign adnexal epithelial proliferations - Follicular:Trichofolliculoma"
    )
    benign_benign_adnexal_epithelial_proliferations_follicular_tricholemmoma = (
        "Benign:Benign adnexal epithelial proliferations - Follicular:Tricholemmoma"
    )
    benign_benign_adnexal_epithelial_proliferations_follicular_tumor_of_follicular_infundibulum = "Benign:Benign adnexal epithelial proliferations - Follicular:Tumor of follicular infundibulum"
    benign_benign_adnexal_epithelial_proliferations_follicular_warty_dyskeratoma = (
        "Benign:Benign adnexal epithelial proliferations - Follicular:Warty dyskeratoma"
    )
    benign_benign_adnexal_epithelial_proliferations_sebaceous_fibrofolliculoma = (
        "Benign:Benign adnexal epithelial proliferations - Sebaceous:Fibrofolliculoma"
    )
    benign_benign_adnexal_epithelial_proliferations_sebaceous_fordyce_spots = (
        "Benign:Benign adnexal epithelial proliferations - Sebaceous:Fordyce spots"
    )
    benign_benign_adnexal_epithelial_proliferations_sebaceous_nevus_sebaceus = (
        "Benign:Benign adnexal epithelial proliferations - Sebaceous:Nevus sebaceus"
    )
    benign_benign_adnexal_epithelial_proliferations_sebaceous_sebaceoma = (
        "Benign:Benign adnexal epithelial proliferations - Sebaceous:Sebaceoma"
    )
    benign_benign_adnexal_epithelial_proliferations_sebaceous_sebaceous_adenoma = (
        "Benign:Benign adnexal epithelial proliferations - Sebaceous:Sebaceous adenoma"
    )
    benign_benign_adnexal_epithelial_proliferations_sebaceous_sebaceous_hyperplasia = (
        "Benign:Benign adnexal epithelial proliferations - Sebaceous:Sebaceous hyperplasia"
    )
    benign_benign_adnexal_epithelial_proliferations_sebaceous_trichodiscoma = (
        "Benign:Benign adnexal epithelial proliferations - Sebaceous:Trichodiscoma"
    )
    benign_benign_epidermal_proliferations_acantholytic_acanthoma = (
        "Benign:Benign epidermal proliferations:Acantholytic acanthoma"
    )
    benign_benign_epidermal_proliferations_clear_cell_acanthoma = (
        "Benign:Benign epidermal proliferations:Clear cell acanthoma"
    )
    benign_benign_epidermal_proliferations_epidermal_nevus = (
        "Benign:Benign epidermal proliferations:Epidermal nevus"
    )
    benign_benign_epidermal_proliferations_epidermolytic_acanthoma = (
        "Benign:Benign epidermal proliferations:Epidermolytic acanthoma"
    )
    benign_benign_epidermal_proliferations_large_cell_acanthoma = (
        "Benign:Benign epidermal proliferations:Large cell acanthoma"
    )
    benign_benign_epidermal_proliferations_lichen_planus_like_keratosis = (
        "Benign:Benign epidermal proliferations:Lichen planus like keratosis"
    )
    benign_benign_epidermal_proliferations_melanoacanthoma = (
        "Benign:Benign epidermal proliferations:Melanoacanthoma"
    )
    benign_benign_epidermal_proliferations_pigmented_benign_keratosis = (
        "Benign:Benign epidermal proliferations:Pigmented benign keratosis"
    )
    benign_benign_epidermal_proliferations_porokeratosis = (
        "Benign:Benign epidermal proliferations:Porokeratosis"
    )
    benign_benign_epidermal_proliferations_seborrheic_keratosis = (
        "Benign:Benign epidermal proliferations:Seborrheic keratosis"
    )
    benign_benign_epidermal_proliferations_solar_lentigo = (
        "Benign:Benign epidermal proliferations:Solar lentigo"
    )
    benign_benign_melanocytic_proliferations_nevus = (
        "Benign:Benign melanocytic proliferations:Nevus"
    )
    benign_benign_melanocytic_proliferations_dermal_melanocytosis = (
        "Benign:Benign melanocytic proliferations:Dermal melanocytosis"
    )
    benign_benign_melanocytic_proliferations_lentiginous_melanocytic_proliferation = (
        "Benign:Benign melanocytic proliferations:Lentiginous melanocytic proliferation"
    )
    benign_benign_melanocytic_proliferations_lentigo_simplex = (
        "Benign:Benign melanocytic proliferations:Lentigo simplex"
    )
    benign_benign_melanocytic_proliferations_pigmented_epithelioid_melanocytoma = (
        "Benign:Benign melanocytic proliferations:Pigmented epithelioid melanocytoma"
    )
    benign_benign_melanocytic_proliferations_proliferative_nodule_in_congenital_melanocytic_nevi_without_atypia = "Benign:Benign melanocytic proliferations:Proliferative nodule in congenital melanocytic nevi without atypia"
    benign_benign_soft_tissue_proliferations_adipocytic_angiolipoma = (
        "Benign:Benign soft tissue proliferations - Adipocytic:Angiolipoma"
    )
    benign_benign_soft_tissue_proliferations_adipocytic_fibrolipoma = (
        "Benign:Benign soft tissue proliferations - Adipocytic:Fibrolipoma"
    )
    benign_benign_soft_tissue_proliferations_adipocytic_lipoma = (
        "Benign:Benign soft tissue proliferations - Adipocytic:Lipoma"
    )
    benign_benign_soft_tissue_proliferations_adipocytic_lipomatous_nevus = (
        "Benign:Benign soft tissue proliferations - Adipocytic:Lipomatous nevus"
    )
    benign_benign_soft_tissue_proliferations_cartilagenous_and_ossifying_accessory_tragus = (
        "Benign:Benign soft tissue proliferations - Cartilagenous and ossifying:Accessory tragus"
    )
    benign_benign_soft_tissue_proliferations_cartilagenous_and_ossifying_extraskeletal_chondroma = "Benign:Benign soft tissue proliferations - Cartilagenous and ossifying:Extraskeletal chondroma"
    benign_benign_soft_tissue_proliferations_cartilagenous_and_ossifying_osteoma_cutis = (
        "Benign:Benign soft tissue proliferations - Cartilagenous and ossifying:Osteoma cutis"
    )
    benign_benign_soft_tissue_proliferations_cartilagenous_and_ossifying_subungual_osteochodroma = "Benign:Benign soft tissue proliferations - Cartilagenous and ossifying:Subungual osteochodroma"
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_angiofibroma = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Angiofibroma"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_dermatofibroma = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Dermatofibroma"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_fibroepithelial_polyp = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Fibroepithelial polyp"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_fibroma = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Fibroma"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_giant_cell_tumor_of_the_tendon_sheath = "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Giant cell tumor of the tendon sheath"
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_juvenile_xanthogranuloma = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Juvenile xanthogranuloma"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_cutaneous_myxoma = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Cutaneous Myxoma"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_non_langerhans_histiocytosis = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Non-Langerhans histiocytosis"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_reticulohistiocytosis = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Reticulohistiocytosis"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_rosai_dorfman_disease = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Rosai-Dorfman disease"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_scar = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Scar"
    )
    benign_benign_soft_tissue_proliferations_muscle_tissue_or_myofibroblastic_angioleiomyoma = (
        "Benign:Benign soft tissue proliferations - Muscle tissue or myofibroblastic:Angioleiomyoma"
    )
    benign_benign_soft_tissue_proliferations_muscle_tissue_or_myofibroblastic_dartoic_muscle_leiomyoma = "Benign:Benign soft tissue proliferations - Muscle tissue or myofibroblastic:Dartoic muscle leiomyoma"
    benign_benign_soft_tissue_proliferations_muscle_tissue_or_myofibroblastic_dermatomyofibroma = "Benign:Benign soft tissue proliferations - Muscle tissue or myofibroblastic:Dermatomyofibroma"
    benign_benign_soft_tissue_proliferations_muscle_tissue_or_myofibroblastic_nodular_fasciitis = "Benign:Benign soft tissue proliferations - Muscle tissue or myofibroblastic:Nodular Fasciitis"
    benign_benign_soft_tissue_proliferations_muscle_tissue_or_myofibroblastic_piloleiomyoma = (
        "Benign:Benign soft tissue proliferations - Muscle tissue or myofibroblastic:Piloleiomyoma"
    )
    benign_benign_soft_tissue_proliferations_muscle_tissue_or_myofibroblastic_smooth_muscle_hamartoma = "Benign:Benign soft tissue proliferations - Muscle tissue or myofibroblastic:Smooth muscle hamartoma"
    benign_benign_soft_tissue_proliferations_myoepithelial_myoepithelioma = (
        "Benign:Benign soft tissue proliferations - Myoepithelial:Myoepithelioma"
    )
    benign_benign_soft_tissue_proliferations_neural_granular_cell_tumor = (
        "Benign:Benign soft tissue proliferations - Neural:Granular cell tumor"
    )
    benign_benign_soft_tissue_proliferations_neural_nerve_sheath_myxoma = (
        "Benign:Benign soft tissue proliferations - Neural:Nerve sheath myxoma"
    )
    benign_benign_soft_tissue_proliferations_neural_neurofibroma = (
        "Benign:Benign soft tissue proliferations - Neural:Neurofibroma"
    )
    benign_benign_soft_tissue_proliferations_neural_plexiform_neurofibroma = (
        "Benign:Benign soft tissue proliferations - Neural:Plexiform Neurofibroma"
    )
    benign_benign_soft_tissue_proliferations_neural_neuroma = (
        "Benign:Benign soft tissue proliferations - Neural:Neuroma"
    )
    benign_benign_soft_tissue_proliferations_neural_perineurioma = (
        "Benign:Benign soft tissue proliferations - Neural:Perineurioma"
    )
    benign_benign_soft_tissue_proliferations_neural_schwannoma = (
        "Benign:Benign soft tissue proliferations - Neural:Schwannoma"
    )
    benign_benign_soft_tissue_proliferations_vascular_acquired_elastotic_hemangioma = (
        "Benign:Benign soft tissue proliferations - Vascular:Acquired elastotic hemangioma"
    )
    benign_benign_soft_tissue_proliferations_vascular_acroangiodermatitis_of_mali = (
        "Benign:Benign soft tissue proliferations - Vascular:Acroangiodermatitis of Mali"
    )
    benign_benign_soft_tissue_proliferations_vascular_angiokeratoma = (
        "Benign:Benign soft tissue proliferations - Vascular:Angiokeratoma"
    )
    benign_benign_soft_tissue_proliferations_vascular_angiolymphoid_hyperplasia_with_eosinophilia = "Benign:Benign soft tissue proliferations - Vascular:Angiolymphoid hyperplasia with eosinophilia"
    benign_benign_soft_tissue_proliferations_vascular_arterio_venous_malformation = (
        "Benign:Benign soft tissue proliferations - Vascular:Arterio-venous malformation"
    )
    benign_benign_soft_tissue_proliferations_vascular_capillary_vascular_malformation = (
        "Benign:Benign soft tissue proliferations - Vascular:Capillary vascular malformation"
    )
    benign_benign_soft_tissue_proliferations_vascular_glomangiomyoma = (
        "Benign:Benign soft tissue proliferations - Vascular:Glomangiomyoma"
    )
    benign_benign_soft_tissue_proliferations_vascular_glomeruloid_hemangioma = (
        "Benign:Benign soft tissue proliferations - Vascular:Glomeruloid hemangioma"
    )
    benign_benign_soft_tissue_proliferations_vascular_glomus_tumor = (
        "Benign:Benign soft tissue proliferations - Vascular:Glomus tumor"
    )
    benign_benign_soft_tissue_proliferations_vascular_hemangioma = (
        "Benign:Benign soft tissue proliferations - Vascular:Hemangioma"
    )
    benign_benign_soft_tissue_proliferations_vascular_lymphangioma = (
        "Benign:Benign soft tissue proliferations - Vascular:Lymphangioma"
    )
    benign_benign_soft_tissue_proliferations_vascular_nevus_anemicus = (
        "Benign:Benign soft tissue proliferations - Vascular:Nevus anemicus"
    )
    benign_benign_soft_tissue_proliferations_vascular_noninvoluting_congenital_hemangioma = (
        "Benign:Benign soft tissue proliferations - Vascular:Noninvoluting congenital hemangioma"
    )
    benign_benign_soft_tissue_proliferations_vascular_other_vascular_or_lymphatic_malformation_or_hamartoma = "Benign:Benign soft tissue proliferations - Vascular:Other vascular or lymphatic malformation or hamartoma"
    benign_benign_soft_tissue_proliferations_vascular_pyogenic_granuloma = (
        "Benign:Benign soft tissue proliferations - Vascular:Pyogenic granuloma"
    )
    benign_benign_soft_tissue_proliferations_vascular_rapidly_involuting_congenital_hemangioma = "Benign:Benign soft tissue proliferations - Vascular:Rapidly involuting congenital hemangioma"
    benign_benign_soft_tissue_proliferations_vascular_telangiectasia = (
        "Benign:Benign soft tissue proliferations - Vascular:Telangiectasia"
    )
    benign_benign_soft_tissue_proliferations_vascular_vascular_spider = (
        "Benign:Benign soft tissue proliferations - Vascular:Vascular spider"
    )
    benign_benign_soft_tissue_proliferations_vascular_venous_lake = (
        "Benign:Benign soft tissue proliferations - Vascular:Venous lake"
    )
    benign_benign_soft_tissue_proliferations_vascular_venous_malformation = (
        "Benign:Benign soft tissue proliferations - Vascular:Venous malformation"
    )
    benign_benign_soft_tissue_proliferations_vascular_verrucous_hemangioma = (
        "Benign:Benign soft tissue proliferations - Vascular:Verrucous hemangioma"
    )
    benign_cysts_comedo = "Benign:Cysts:Comedo"
    benign_cysts_digital_mucous_cyst = "Benign:Cysts:Digital mucous cyst"
    benign_cysts_dilated_pore = "Benign:Cysts:Dilated pore"
    benign_cysts_infundibular_or_epidermal_cyst = "Benign:Cysts:Infundibular or epidermal cyst"
    benign_cysts_sebaceous_cyst = "Benign:Cysts:Sebaceous cyst"
    benign_cysts_keratinous_cyst = "Benign:Cysts:Keratinous cyst"
    benign_cysts_milium = "Benign:Cysts:Milium"
    benign_cysts_steatocystoma = "Benign:Cysts:Steatocystoma"
    benign_cysts_trichilemmal_or_isthmic_catagen_or_pilar_cyst = (
        "Benign:Cysts:Trichilemmal or isthmic-catagen or pilar cyst"
    )
    benign_exogenous_foreign_body_granuloma = "Benign:Exogenous:Foreign body granuloma"
    benign_exogenous_tattoo = "Benign:Exogenous:Tattoo"
    benign_flat_melanotic_pigmentations_not_melanocytic_nevus_cafe_au_lait_macule_or_patch = (
        "Benign:Flat melanotic pigmentations - not melanocytic nevus:Cafe au lait macule or patch"
    )
    benign_flat_melanotic_pigmentations_not_melanocytic_nevus_ephelis = (
        "Benign:Flat melanotic pigmentations - not melanocytic nevus:Ephelis"
    )
    benign_flat_melanotic_pigmentations_not_melanocytic_nevus_ink_spot_lentigo = (
        "Benign:Flat melanotic pigmentations - not melanocytic nevus:Ink-spot lentigo"
    )
    benign_flat_melanotic_pigmentations_not_melanocytic_nevus_lentigo_nos = (
        "Benign:Flat melanotic pigmentations - not melanocytic nevus:Lentigo NOS"
    )
    benign_flat_melanotic_pigmentations_not_melanocytic_nevus_mucosal_melanotic_macule = (
        "Benign:Flat melanotic pigmentations - not melanocytic nevus:Mucosal melanotic macule"
    )
    benign_hemorrhagic_lesions_hemorrhage = "Benign:Hemorrhagic lesions:Hemorrhage"
    benign_inflammatory_or_infectious_diseases_verruca = (
        "Benign:Inflammatory or infectious diseases:Verruca"
    )
    benign_inflammatory_or_infectious_diseases_molluscum = (
        "Benign:Inflammatory or infectious diseases:Molluscum"
    )
    benign_langerhans_cell_proliferations_erdheim_chester_disease = (
        "Benign:Langerhans cell proliferations:Erdheim Chester disease"
    )
    benign_langerhans_cell_proliferations_indeterminate_cell_histiocytosis = (
        "Benign:Langerhans cell proliferations:Indeterminate cell histiocytosis"
    )
    benign_langerhans_cell_proliferations_langerhans_cell_histiocytosis = (
        "Benign:Langerhans cell proliferations:Langerhans cell histiocytosis"
    )
    benign_langerhans_cell_proliferations_mixed_langerhans_cell_histiocytosis_and_erdheim_chester_disease = "Benign:Langerhans cell proliferations:Mixed Langerhans cell histiocytosis and Erdheim Chester disease"
    benign_mast_cell_proliferations_maculopapular_mastocytoma = (
        "Benign:Mast cell proliferations:Maculopapular mastocytoma"
    )
    benign_mast_cell_proliferations_mastocytoma_solitary_or_unifocal = (
        "Benign:Mast cell proliferations:Mastocytoma, Solitary or unifocal"
    )
    benign_mast_cell_proliferations_mastocytosis = "Benign:Mast cell proliferations:Mastocytosis"
    indeterminate_indeterminate_epidermal_proliferations_solar_or_actinic_cheilitis = (
        "Indeterminate:Indeterminate epidermal proliferations:Solar or actinic cheilitis"
    )
    indeterminate_indeterminate_epidermal_proliferations_solar_or_actinic_keratosis = (
        "Indeterminate:Indeterminate epidermal proliferations:Solar or actinic keratosis"
    )
    indeterminate_indeterminate_melanocytic_proliferations_atypical_spitz_tumor = (
        "Indeterminate:Indeterminate melanocytic proliferations:Atypical Spitz tumor"
    )
    indeterminate_indeterminate_melanocytic_proliferations_atypical_intraepithelial_melanocytic_proliferation = "Indeterminate:Indeterminate melanocytic proliferations:Atypical intraepithelial melanocytic proliferation"
    indeterminate_indeterminate_melanocytic_proliferations_atypical_melanocytic_neoplasm = (
        "Indeterminate:Indeterminate melanocytic proliferations:Atypical melanocytic neoplasm"
    )
    indeterminate_indeterminate_melanocytic_proliferations_atypical_pigmented_spindle_cell_tumor = "Indeterminate:Indeterminate melanocytic proliferations:Atypical pigmented spindle cell tumor"
    indeterminate_indeterminate_melanocytic_proliferations_atypical_proliferative_nodules_in_congenital_melanocytic_nevus = "Indeterminate:Indeterminate melanocytic proliferations:Atypical proliferative nodules in congenital melanocytic nevus"
    indeterminate_indeterminate_melanocytic_proliferations_melanocytic_tumor_of_uncertain_malignant_potential = "Indeterminate:Indeterminate melanocytic proliferations:Melanocytic tumor of uncertain malignant potential"
    indeterminate_indeterminate_melanocytic_proliferations_superficial_atypical_melanocytic_proliferation_of_uncertain_significance = "Indeterminate:Indeterminate melanocytic proliferations:Superficial atypical melanocytic proliferation of uncertain significance"
    malignant_lymphocytic_proliferations_b_cell_ebv_positive_mucocutaneous_ulcer = (
        "Malignant:Lymphocytic proliferations - B-Cell:EBV positive mucocutaneous ulcer"
    )
    malignant_lymphocytic_proliferations_b_cell_intravascular_large_b_cell_lymphoma = (
        "Malignant:Lymphocytic proliferations - B-Cell:Intravascular large B-cell lymphoma"
    )
    malignant_lymphocytic_proliferations_b_cell_lymphocytic_proliferation_b_cell_other = (
        "Malignant:Lymphocytic proliferations - B-Cell:Lymphocytic proliferation, B-Cell, other"
    )
    malignant_lymphocytic_proliferations_b_cell_primary_cutaneous_follicle_center_lymphoma = (
        "Malignant:Lymphocytic proliferations - B-Cell:Primary cutaneous follicle center lymphoma"
    )
    malignant_lymphocytic_proliferations_b_cell_primary_cutaneous_large_b_cell_lymphoma = (
        "Malignant:Lymphocytic proliferations - B-Cell:Primary cutaneous large B-Cell lymphoma"
    )
    malignant_lymphocytic_proliferations_b_cell_primary_cutaneous_marginal_zone_lymphoma = (
        "Malignant:Lymphocytic proliferations - B-Cell:Primary cutaneous marginal zone lymphoma"
    )
    malignant_lymphocytic_proliferations_t_cellnk_adult_t_cell_leukemia_or_lymphoma = (
        "Malignant:Lymphocytic proliferations - T-Cell/NK:Adult T-cell leukemia or lymphoma"
    )
    malignant_lymphocytic_proliferations_t_cellnk_chronic_active_ebv_infection = (
        "Malignant:Lymphocytic proliferations - T-Cell/NK:Chronic active EBV infection"
    )
    malignant_lymphocytic_proliferations_t_cellnk_extranodal_t_cellnk_lymphoma = (
        "Malignant:Lymphocytic proliferations - T-Cell/NK:Extranodal T-cell/NK lymphoma"
    )
    malignant_lymphocytic_proliferations_t_cellnk_lymphocytic_proliferation_t_cellnk = (
        "Malignant:Lymphocytic proliferations - T-Cell/NK:Lymphocytic proliferation, T-Cell/NK"
    )
    malignant_lymphocytic_proliferations_t_cellnk_mycosis_fungoides = (
        "Malignant:Lymphocytic proliferations - T-Cell/NK:Mycosis fungoides"
    )
    malignant_lymphocytic_proliferations_t_cellnk_primary_cutaneous_cd30_lymphoproliferative_disease = "Malignant:Lymphocytic proliferations - T-Cell/NK:Primary cutaneous CD30+ lymphoproliferative disease"
    malignant_lymphocytic_proliferations_t_cellnk_primary_cutaneous_cd4_small_or_medium_t_cell_lymphoproliferative_disorder = "Malignant:Lymphocytic proliferations - T-Cell/NK:Primary cutaneous CD4+ small or medium T-cell lymphoproliferative disorder"
    malignant_lymphocytic_proliferations_t_cellnk_primary_cutaneous_peripheral_t_cell_lymphoma = "Malignant:Lymphocytic proliferations - T-Cell/NK:Primary cutaneous peripheral T-cell lymphoma"
    malignant_lymphocytic_proliferations_t_cellnk_sezary_syndrome = (
        "Malignant:Lymphocytic proliferations - T-Cell/NK:Sezary syndrome"
    )
    malignant_lymphocytic_proliferations_t_cellnk_subcutaneous_panniculitis_like_t_cell_lymphoma = "Malignant:Lymphocytic proliferations - T-Cell/NK:Subcutaneous panniculitis-like T-cell lymphoma"
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine_adenoid_cystic_carcinoma = "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine:Adenoid cystic carcinoma"
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine_adnexal_adenocarcinoma_arising_in_association_with_spiradenoma_cylindroma_or_spiradenocylindroma = "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine:Adnexal adenocarcinoma arising in association with spiradenoma, cylindroma, or spiradenocylindroma"
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine_apocrine_carcinoma = "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine:Apocrine carcinoma"
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine_digital_papillary_carcinoma = "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine:Digital papillary carcinoma"
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine_hidradenocarcinoma = "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine:Hidradenocarcinoma"
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine_malignant_mixed_tumor = "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine:Malignant mixed tumor"
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine_microcystic_adnexal_carcinoma = "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine:Microcystic adnexal carcinoma"
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine_mucinous_carcinoma = "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine:Mucinous carcinoma"
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine_paget_disease = (
        "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine:Paget disease"
    )
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine_porocarcinoma = (
        "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine:Porocarcinoma"
    )
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine_tubular_carcinoma = "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine:Tubular carcinoma"
    malignant_malignant_adnexal_epithelial_proliferations_follicular_basal_cell_carcinoma = (
        "Malignant:Malignant adnexal epithelial proliferations - Follicular:Basal cell carcinoma"
    )
    malignant_malignant_adnexal_epithelial_proliferations_follicular_baso_squamous_carcinoma = (
        "Malignant:Malignant adnexal epithelial proliferations - Follicular:Baso-squamous carcinoma"
    )
    malignant_malignant_adnexal_epithelial_proliferations_follicular_matrical_or_pilomatrical_carcinoma = "Malignant:Malignant adnexal epithelial proliferations - Follicular:Matrical or pilomatrical carcinoma"
    malignant_malignant_adnexal_epithelial_proliferations_follicular_proliferating_trichilemmal_carcinoma = "Malignant:Malignant adnexal epithelial proliferations - Follicular:Proliferating trichilemmal carcinoma"
    malignant_malignant_adnexal_epithelial_proliferations_sebaceous_sebaceous_carcinoma = (
        "Malignant:Malignant adnexal epithelial proliferations - Sebaceous:Sebaceous carcinoma"
    )
    malignant_malignant_epidermal_proliferations_bowenoid_papulosis = (
        "Malignant:Malignant epidermal proliferations:Bowenoid papulosis"
    )
    malignant_malignant_epidermal_proliferations_keratoacanthoma = (
        "Malignant:Malignant epidermal proliferations:Keratoacanthoma"
    )
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_in_situ = (
        "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma in situ"
    )
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive = (
        "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive"
    )
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_nos = (
        "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, NOS"
    )
    malignant_malignant_epidermal_proliferations_verrucous_carcinoma = (
        "Malignant:Malignant epidermal proliferations:Verrucous carcinoma"
    )
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive = (
        "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive"
    )
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_in_situ = (
        "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma in situ"
    )
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_metastasis = (
        "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma metastasis"
    )
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_nos = (
        "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma, NOS"
    )
    malignant_malignant_soft_tissue_proliferations_adipocytic_liposarcoma = (
        "Malignant:Malignant soft tissue proliferations - Adipocytic:Liposarcoma"
    )
    malignant_malignant_soft_tissue_proliferations_cartilagenous_and_ossifying_extraskeletal_osteosarcoma = "Malignant:Malignant soft tissue proliferations - Cartilagenous and ossifying:Extraskeletal osteosarcoma"
    malignant_malignant_soft_tissue_proliferations_fibro_histiocytic_atypical_fibroxanthoma = (
        "Malignant:Malignant soft tissue proliferations - Fibro-histiocytic:Atypical fibroxanthoma"
    )
    malignant_malignant_soft_tissue_proliferations_fibro_histiocytic_dermatofibrosarcoma_protuberans = "Malignant:Malignant soft tissue proliferations - Fibro-histiocytic:Dermatofibrosarcoma protuberans"
    malignant_malignant_soft_tissue_proliferations_fibro_histiocytic_epithelioid_sarcoma = (
        "Malignant:Malignant soft tissue proliferations - Fibro-histiocytic:Epithelioid sarcoma"
    )
    malignant_malignant_soft_tissue_proliferations_fibro_histiocytic_fibrosarcoma = (
        "Malignant:Malignant soft tissue proliferations - Fibro-histiocytic:Fibrosarcoma"
    )
    malignant_malignant_soft_tissue_proliferations_fibro_histiocytic_pleomorphic_undifferntiated_sarcoma = "Malignant:Malignant soft tissue proliferations - Fibro-histiocytic:Pleomorphic undifferntiated sarcoma"
    malignant_malignant_soft_tissue_proliferations_muscle_tissue_or_myofibroblastic_atypical_intradermal_smooth_muscle_tumor = "Malignant:Malignant soft tissue proliferations - Muscle tissue or myofibroblastic:Atypical intradermal smooth muscle tumor"
    malignant_malignant_soft_tissue_proliferations_muscle_tissue_or_myofibroblastic_leiomyosarcoma_cutaneous = "Malignant:Malignant soft tissue proliferations - Muscle tissue or myofibroblastic:Leiomyosarcoma, Cutaneous"
    malignant_malignant_soft_tissue_proliferations_muscle_tissue_or_myofibroblastic_rhabdomyoscaroma_cutaneous = "Malignant:Malignant soft tissue proliferations - Muscle tissue or myofibroblastic:Rhabdomyoscaroma, Cutaneous"
    malignant_malignant_soft_tissue_proliferations_myoepithelial_myoepithelial_sarcoma = (
        "Malignant:Malignant soft tissue proliferations - Myoepithelial:Myoepithelial sarcoma"
    )
    malignant_malignant_soft_tissue_proliferations_neural_malignant_granular_cell_tumor = (
        "Malignant:Malignant soft tissue proliferations - Neural:Malignant granular cell tumor"
    )
    malignant_malignant_soft_tissue_proliferations_neural_malignant_peripheral_nerve_sheath_tumor = "Malignant:Malignant soft tissue proliferations - Neural:Malignant peripheral nerve sheath tumor"
    malignant_malignant_soft_tissue_proliferations_unknown_or_other_histiogenesis_ewing_sarcoma_primary_cutaenous = "Malignant:Malignant soft tissue proliferations - Unknown or other histiogenesis:Ewing sarcoma, Primary cutaenous"
    malignant_malignant_soft_tissue_proliferations_vascular_angiosarcoma_cutaneous = (
        "Malignant:Malignant soft tissue proliferations - Vascular:Angiosarcoma cutaneous"
    )
    malignant_malignant_soft_tissue_proliferations_vascular_hemangioendothelioma = (
        "Malignant:Malignant soft tissue proliferations - Vascular:Hemangioendothelioma"
    )
    malignant_malignant_soft_tissue_proliferations_vascular_kaposi_sarcoma = (
        "Malignant:Malignant soft tissue proliferations - Vascular:Kaposi sarcoma"
    )
    malignant_malignant_soft_tissue_proliferations_vascular_malignant_glomus_tumor = (
        "Malignant:Malignant soft tissue proliferations - Vascular:Malignant glomus tumor"
    )
    malignant_merkel_cell_proliferation_merkel_cell_carcinoma = (
        "Malignant:Merkel cell proliferation:Merkel cell carcinoma"
    )
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_hidradenoma_apocrine_hidradenoma_apocrine_predominantly_with_clear_cells = "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Hidradenoma, Apocrine:Hidradenoma, Apocrine, Predominantly with clear cells"
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_hidradenoma_hidradenoma_poroid = "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Hidradenoma:Hidradenoma, Poroid"
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_mixed_tumor_mixed_tumor_apocrine_type = "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Mixed tumor:Mixed tumor, Apocrine type"
    benign_benign_adnexal_epithelial_proliferations_apocrine_or_eccrine_mixed_tumor_mixed_tumor_eccrine_type = "Benign:Benign adnexal epithelial proliferations - Apocrine or Eccrine:Mixed tumor:Mixed tumor, Eccrine type"
    benign_benign_adnexal_epithelial_proliferations_follicular_trichoepithelioma_trichoepithelioma_desmoplastic = "Benign:Benign adnexal epithelial proliferations - Follicular:Trichoepithelioma:Trichoepithelioma, Desmoplastic"
    benign_benign_adnexal_epithelial_proliferations_follicular_tricholemmoma_tricholemmoma_desmoplastic = "Benign:Benign adnexal epithelial proliferations - Follicular:Tricholemmoma:Tricholemmoma, Desmoplastic"
    benign_benign_epidermal_proliferations_seborrheic_keratosis_seborrheic_keratosis_clonal = (
        "Benign:Benign epidermal proliferations:Seborrheic keratosis:Seborrheic keratosis, Clonal"
    )
    benign_benign_melanocytic_proliferations_nevus_blue_nevus = (
        "Benign:Benign melanocytic proliferations:Nevus:Blue nevus"
    )
    benign_benign_melanocytic_proliferations_dermal_melanocytosis_mongolian_spot = (
        "Benign:Benign melanocytic proliferations:Dermal melanocytosis:Mongolian spot"
    )
    benign_benign_melanocytic_proliferations_dermal_melanocytosis_nevus_of_ito = (
        "Benign:Benign melanocytic proliferations:Dermal melanocytosis:Nevus of Ito"
    )
    benign_benign_melanocytic_proliferations_dermal_melanocytosis_nevus_of_ota = (
        "Benign:Benign melanocytic proliferations:Dermal melanocytosis:Nevus of Ota"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_acral = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Acral"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_agminated = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Agminated"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_atypical_dysplastic_or_clark = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Atypical, Dysplastic, or Clark"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_bap_1_deficient = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, BAP-1 deficient"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_balloon_cell = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Balloon cell"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_combined = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Combined"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_congenital = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Congenital"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_deep_penetrating = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Deep penetrating"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_halo = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Halo"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_lentiginous = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Lentiginous"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_meyerson = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Meyerson"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_nos_compound = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, NOS, Compound"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_nos_dermal = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, NOS, Dermal"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_nos_junctional = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, NOS, Junctional"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_of_special_anatomic_site = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Of special anatomic site"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_recurrent_or_persistent = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Recurrent or persistent"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_reed = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Reed"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_spilus = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Spilus"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_spitz = (
        "Benign:Benign melanocytic proliferations:Nevus:Nevus, Spitz"
    )
    benign_benign_soft_tissue_proliferations_adipocytic_lipoma_lipoma_spindle_cell = (
        "Benign:Benign soft tissue proliferations - Adipocytic:Lipoma:Lipoma, Spindle cell"
    )
    benign_benign_soft_tissue_proliferations_adipocytic_lipoma_lipoma_pleomorphic = (
        "Benign:Benign soft tissue proliferations - Adipocytic:Lipoma:Lipoma, Pleomorphic"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_angiofibroma_angiofibroma_facial = "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Angiofibroma:Angiofibroma, Facial"
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_angiofibroma_angiofibroma_penile = "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Angiofibroma:Angiofibroma, Penile"
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_angiofibroma_angiofibroma_periungual = "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Angiofibroma:Angiofibroma, Periungual"
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_dermatofibroma_dermatofibroma_aneurysmal = "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Dermatofibroma:Dermatofibroma, Aneurysmal"
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_dermatofibroma_dermatofibroma_atypical = "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Dermatofibroma:Dermatofibroma, Atypical"
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_dermatofibroma_dermatofibroma_cellular = "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Dermatofibroma:Dermatofibroma, Cellular"
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_dermatofibroma_dermatofibroma_epithelioid = "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Dermatofibroma:Dermatofibroma, Epithelioid"
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_dermatofibroma_dermatofibroma_hemosiderotic = "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Dermatofibroma:Dermatofibroma, Hemosiderotic"
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_fibroma_fibroma_pleomorphic = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Fibroma:Fibroma, Pleomorphic"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_fibroma_fibroma_sclerotic = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Fibroma:Fibroma, Sclerotic"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_scar_scar_hypertrophic = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Scar:Scar, Hypertrophic"
    )
    benign_benign_soft_tissue_proliferations_fibro_histiocytic_scar_scar_keloid = (
        "Benign:Benign soft tissue proliferations - Fibro-histiocytic:Scar:Scar, Keloid"
    )
    benign_benign_soft_tissue_proliferations_neural_granular_cell_tumor_granular_cell_tumor_neural_and_s100_positive = "Benign:Benign soft tissue proliferations - Neural:Granular cell tumor:Granular cell tumor, neural and s100 positive"
    benign_benign_soft_tissue_proliferations_neural_granular_cell_tumor_granular_cell_tumor_non_neural_and_s100_negative = "Benign:Benign soft tissue proliferations - Neural:Granular cell tumor:Granular cell tumor, non-neural and s100 negative"
    benign_benign_soft_tissue_proliferations_neural_neuroma_neuroma_palisaded_and_encapsulated = "Benign:Benign soft tissue proliferations - Neural:Neuroma:Neuroma, Palisaded and encapsulated"
    benign_benign_soft_tissue_proliferations_neural_neuroma_neuroma_traumatic = (
        "Benign:Benign soft tissue proliferations - Neural:Neuroma:Neuroma, Traumatic"
    )
    benign_benign_soft_tissue_proliferations_vascular_hemangioma_hemangioma_cherry = (
        "Benign:Benign soft tissue proliferations - Vascular:Hemangioma:Hemangioma, Cherry"
    )
    benign_benign_soft_tissue_proliferations_vascular_hemangioma_hemangioma_hobnail = (
        "Benign:Benign soft tissue proliferations - Vascular:Hemangioma:Hemangioma, Hobnail"
    )
    benign_benign_soft_tissue_proliferations_vascular_hemangioma_hemangioma_infantile = (
        "Benign:Benign soft tissue proliferations - Vascular:Hemangioma:Hemangioma, Infantile"
    )
    benign_benign_soft_tissue_proliferations_vascular_hemangioma_hemangioma_tufted = (
        "Benign:Benign soft tissue proliferations - Vascular:Hemangioma:Hemangioma, Tufted"
    )
    benign_benign_soft_tissue_proliferations_vascular_lymphangioma_lymphangioma_superficial = (
        "Benign:Benign soft tissue proliferations - Vascular:Lymphangioma:Lymphangioma, superficial"
    )
    benign_benign_soft_tissue_proliferations_vascular_lymphangioma_lymphangioma_deep = (
        "Benign:Benign soft tissue proliferations - Vascular:Lymphangioma:Lymphangioma, deep"
    )
    benign_cysts_sebaceous_cyst_infundibular_sebaceous = (
        "Benign:Cysts:Sebaceous cyst:Infundibular, Sebaceous"
    )
    benign_cysts_sebaceous_cyst_epidermal_sebaceous = (
        "Benign:Cysts:Sebaceous cyst:Epidermal, Sebaceous"
    )
    benign_cysts_keratinous_cyst_infundibular_keratinous = (
        "Benign:Cysts:Keratinous cyst:Infundibular, Keratinous"
    )
    benign_cysts_keratinous_cyst_epidermal_keratinous = (
        "Benign:Cysts:Keratinous cyst:Epidermal, Keratinous"
    )
    benign_cysts_trichilemmal_or_isthmic_catagen_or_pilar_cyst_trichilemmal_cyst = (
        "Benign:Cysts:Trichilemmal or isthmic-catagen or pilar cyst:Trichilemmal cyst"
    )
    benign_cysts_trichilemmal_or_isthmic_catagen_or_pilar_cyst_isthmic_catagen_cyst = (
        "Benign:Cysts:Trichilemmal or isthmic-catagen or pilar cyst:Isthmic-catagen cyst"
    )
    benign_cysts_trichilemmal_or_isthmic_catagen_or_pilar_cyst_pilar_cyst = (
        "Benign:Cysts:Trichilemmal or isthmic-catagen or pilar cyst:Pilar cyst"
    )
    benign_hemorrhagic_lesions_hemorrhage_dermal_and_subcutaneous_hemorhage = (
        "Benign:Hemorrhagic lesions:Hemorrhage:Dermal and subcutaneous hemorhage"
    )
    benign_hemorrhagic_lesions_hemorrhage_mucosal_hemorrhage = (
        "Benign:Hemorrhagic lesions:Hemorrhage:Mucosal hemorrhage"
    )
    benign_hemorrhagic_lesions_hemorrhage_subcorneal_and_intracorneal_hemorrhage = (
        "Benign:Hemorrhagic lesions:Hemorrhage:Subcorneal and intracorneal hemorrhage"
    )
    benign_hemorrhagic_lesions_hemorrhage_subungual_hemorrhage = (
        "Benign:Hemorrhagic lesions:Hemorrhage:Subungual hemorrhage"
    )
    benign_langerhans_cell_proliferations_langerhans_cell_histiocytosis_langerhans_cell_histiocytosis_diffuse_or_multifocal = "Benign:Langerhans cell proliferations:Langerhans cell histiocytosis:Langerhans cell histiocytosis, Diffuse or multifocal"
    benign_langerhans_cell_proliferations_langerhans_cell_histiocytosis_langerhans_cell_histiocytosis_solitary_or_unifocal = "Benign:Langerhans cell proliferations:Langerhans cell histiocytosis:Langerhans cell histiocytosis, Solitary or unifocal"
    benign_mast_cell_proliferations_mastocytosis_mastocytosis_diffuse_or_multifocal = (
        "Benign:Mast cell proliferations:Mastocytosis:Mastocytosis, Diffuse or multifocal"
    )
    indeterminate_indeterminate_epidermal_proliferations_solar_or_actinic_keratosis_actinic_keratosis_acantholytic = "Indeterminate:Indeterminate epidermal proliferations:Solar or actinic keratosis:Actinic keratosis, Acantholytic"
    indeterminate_indeterminate_epidermal_proliferations_solar_or_actinic_keratosis_actinic_keratosis_atrophic = "Indeterminate:Indeterminate epidermal proliferations:Solar or actinic keratosis:Actinic keratosis, Atrophic"
    indeterminate_indeterminate_epidermal_proliferations_solar_or_actinic_keratosis_actinic_keratosis_bowenoid = "Indeterminate:Indeterminate epidermal proliferations:Solar or actinic keratosis:Actinic keratosis, Bowenoid"
    indeterminate_indeterminate_epidermal_proliferations_solar_or_actinic_keratosis_actinic_keratosis_hypertrophic = "Indeterminate:Indeterminate epidermal proliferations:Solar or actinic keratosis:Actinic keratosis, Hypertrophic"
    indeterminate_indeterminate_epidermal_proliferations_solar_or_actinic_keratosis_actinic_keratosis_lichenoid = "Indeterminate:Indeterminate epidermal proliferations:Solar or actinic keratosis:Actinic keratosis, Lichenoid"
    indeterminate_indeterminate_melanocytic_proliferations_atypical_spitz_tumor_atypical_spitz_tumor_compound = "Indeterminate:Indeterminate melanocytic proliferations:Atypical Spitz tumor:Atypical Spitz tumor, Compound"
    indeterminate_indeterminate_melanocytic_proliferations_atypical_spitz_tumor_atypical_spitz_tumor_dermal = "Indeterminate:Indeterminate melanocytic proliferations:Atypical Spitz tumor:Atypical Spitz tumor, Dermal"
    indeterminate_indeterminate_melanocytic_proliferations_atypical_spitz_tumor_atypical_spitz_tumor_junctional = "Indeterminate:Indeterminate melanocytic proliferations:Atypical Spitz tumor:Atypical Spitz tumor, Junctional"
    malignant_lymphocytic_proliferations_t_cellnk_extranodal_t_cellnk_lymphoma_extranodal_t_cellnk_lymphoma_nasal_type = "Malignant:Lymphocytic proliferations - T-Cell/NK:Extranodal T-cell/NK lymphoma:Extranodal T-cell/NK lymphoma, Nasal type"
    malignant_lymphocytic_proliferations_t_cellnk_mycosis_fungoides_mycosis_fungoides_folliculotropic = "Malignant:Lymphocytic proliferations - T-Cell/NK:Mycosis fungoides:Mycosis fungoides, Folliculotropic"
    malignant_lymphocytic_proliferations_t_cellnk_mycosis_fungoides_mycosis_fungoides_granulomatous_slack_skin = "Malignant:Lymphocytic proliferations - T-Cell/NK:Mycosis fungoides:Mycosis fungoides, Granulomatous slack skin"
    malignant_lymphocytic_proliferations_t_cellnk_mycosis_fungoides_mycosis_fungoides_pagetoid_reticulosis = "Malignant:Lymphocytic proliferations - T-Cell/NK:Mycosis fungoides:Mycosis fungoides, Pagetoid reticulosis"
    malignant_lymphocytic_proliferations_t_cellnk_mycosis_fungoides_mycosis_fungoides_with_large_cell_transformation = "Malignant:Lymphocytic proliferations - T-Cell/NK:Mycosis fungoides:Mycosis fungoides, With large cell transformation"
    malignant_lymphocytic_proliferations_t_cellnk_primary_cutaneous_cd30_lymphoproliferative_disease_cutanous_anaplastic_large_cell_lymphoma = "Malignant:Lymphocytic proliferations - T-Cell/NK:Primary cutaneous CD30+ lymphoproliferative disease:Cutanous anaplastic large cell lymphoma"
    malignant_lymphocytic_proliferations_t_cellnk_primary_cutaneous_cd30_lymphoproliferative_disease_lymphomatoid_papulosis = "Malignant:Lymphocytic proliferations - T-Cell/NK:Primary cutaneous CD30+ lymphoproliferative disease:Lymphomatoid papulosis"
    malignant_lymphocytic_proliferations_t_cellnk_primary_cutaneous_peripheral_t_cell_lymphoma_primary_cutaneous_peripheral_t_cell_lymphoma_rare_subtype = "Malignant:Lymphocytic proliferations - T-Cell/NK:Primary cutaneous peripheral T-cell lymphoma:Primary cutaneous peripheral T-cell lymphoma, Rare subtype"
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine_paget_disease_paget_disease_extra_mammary = "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine:Paget disease:Paget disease, Extra-mammary"
    malignant_malignant_adnexal_epithelial_proliferations_apocrine_or_eccrine_paget_disease_paget_disease_mammary = "Malignant:Malignant adnexal epithelial proliferations - Apocrine or Eccrine:Paget disease:Paget disease, Mammary"
    malignant_malignant_adnexal_epithelial_proliferations_follicular_basal_cell_carcinoma_basal_cell_carcinoma_with_adnexal_differentiation = "Malignant:Malignant adnexal epithelial proliferations - Follicular:Basal cell carcinoma:Basal cell carcinoma with adnexal differentiation"
    malignant_malignant_adnexal_epithelial_proliferations_follicular_basal_cell_carcinoma_basal_cell_carcinoma_with_sarcomatoid_differentiation = "Malignant:Malignant adnexal epithelial proliferations - Follicular:Basal cell carcinoma:Basal cell carcinoma with sarcomatoid differentiation"
    malignant_malignant_adnexal_epithelial_proliferations_follicular_basal_cell_carcinoma_basal_cell_carcinoma_combined_subtypes = "Malignant:Malignant adnexal epithelial proliferations - Follicular:Basal cell carcinoma:Basal cell carcinoma, Combined subtypes"
    malignant_malignant_adnexal_epithelial_proliferations_follicular_basal_cell_carcinoma_basal_cell_carcinoma_fibroeipthelial = "Malignant:Malignant adnexal epithelial proliferations - Follicular:Basal cell carcinoma:Basal cell carcinoma, Fibroeipthelial"
    malignant_malignant_adnexal_epithelial_proliferations_follicular_basal_cell_carcinoma_basal_cell_carcinoma_infiltrating = "Malignant:Malignant adnexal epithelial proliferations - Follicular:Basal cell carcinoma:Basal cell carcinoma, Infiltrating"
    malignant_malignant_adnexal_epithelial_proliferations_follicular_basal_cell_carcinoma_basal_cell_carcinoma_micronodular = "Malignant:Malignant adnexal epithelial proliferations - Follicular:Basal cell carcinoma:Basal cell carcinoma, Micronodular"
    malignant_malignant_adnexal_epithelial_proliferations_follicular_basal_cell_carcinoma_basal_cell_carcinoma_nodular = "Malignant:Malignant adnexal epithelial proliferations - Follicular:Basal cell carcinoma:Basal cell carcinoma, Nodular"
    malignant_malignant_adnexal_epithelial_proliferations_follicular_basal_cell_carcinoma_basal_cell_carcinoma_sclerosing_or_morpheaform = "Malignant:Malignant adnexal epithelial proliferations - Follicular:Basal cell carcinoma:Basal cell carcinoma, Sclerosing or morpheaform"
    malignant_malignant_adnexal_epithelial_proliferations_follicular_basal_cell_carcinoma_basal_cell_carcinoma_superficial = "Malignant:Malignant adnexal epithelial proliferations - Follicular:Basal cell carcinoma:Basal cell carcinoma, Superficial"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_in_situ_squamous_cell_carcinoma_in_situ_bowens_disease = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma in situ:Squamous cell carcinoma in situ, Bowens disease"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_acantholytic = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Acantholytic"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_adeno_squamous = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Adeno-squamous"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_clear_cell = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Clear cell"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_keratoacanthoma_type = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Keratoacanthoma-type"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_sarcomatoid = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Sarcomatoid"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_spindle_cell = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Spindle cell"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_verrucous = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Verrucous"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_nos_well_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, NOS, well differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_nos_moderately_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, NOS, moderately differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_nos_poorly_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, NOS, poorly differentiated"
    malignant_malignant_epidermal_proliferations_verrucous_carcinoma_verrucous_carcinoma_carcinoma_cuniculatum_type = "Malignant:Malignant epidermal proliferations:Verrucous carcinoma:Verrucous carcinoma, Carcinoma cuniculatum type"
    malignant_malignant_epidermal_proliferations_verrucous_carcinoma_verrucous_carcinoma_giant_condyloma_type = "Malignant:Malignant epidermal proliferations:Verrucous carcinoma:Verrucous carcinoma, Giant condyloma type"
    malignant_malignant_epidermal_proliferations_verrucous_carcinoma_verrucous_carcinoma_oral_florid_papilomatosis_type = "Malignant:Malignant epidermal proliferations:Verrucous carcinoma:Verrucous carcinoma, Oral florid papilomatosis type"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_acral_or_acral_lentiginous = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Acral or Acral-lentiginous"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_arising_in_a_congenital_nevus = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Arising in a congenital nevus"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_associated_with_a_nevus = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Associated with a nevus"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_blue_nevus_like = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Blue nevus-like"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_desmoplastic = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Desmoplastic"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_heavily_pigmented = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Heavily pigmented"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_mucosal = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Mucosal"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_neurotropic = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Neurotropic"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_nevoid = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Nevoid"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_nodular = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Nodular"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_on_chronically_sun_exposed_skin_or_lentigo_maligna_melanoma = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, On chronically sun-exposed skin or lentigo maligna melanoma"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_pigmented_spindle_cell_nevus_like = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Pigmented spindle cell nevus like"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_recurrent_or_persistent = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Recurrent or persistent"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_spitzoid = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Spitzoid"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_superficial_spreading = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Superficial spreading"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_in_situ_melanoma_in_situ_acral_or_acral_lentiginous = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma in situ:Melanoma in situ, Acral or acral-lentiginous"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_in_situ_melanoma_in_situ_lentigo_maligna_type = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma in situ:Melanoma in situ, Lentigo maligna type"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_in_situ_melanoma_in_situ_mucosal = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma in situ:Melanoma in situ, Mucosal"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_in_situ_melanoma_in_situ_recurrent_or_persistent = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma in situ:Melanoma in situ, Recurrent or persistent"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_in_situ_melanoma_in_situ_superficial_spreading = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma in situ:Melanoma in situ, Superficial spreading"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_in_situ_melanoma_in_situ_associated_with_a_nevus = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma in situ:Melanoma in situ, associated with a nevus"
    malignant_malignant_soft_tissue_proliferations_adipocytic_liposarcoma_liposarcoma_undifferentiated = "Malignant:Malignant soft tissue proliferations - Adipocytic:Liposarcoma:Liposarcoma, Undifferentiated"
    malignant_malignant_soft_tissue_proliferations_adipocytic_liposarcoma_liposarcoma_well_differentiated = "Malignant:Malignant soft tissue proliferations - Adipocytic:Liposarcoma:Liposarcoma, Well differentiated"
    malignant_malignant_soft_tissue_proliferations_vascular_angiosarcoma_cutaneous_angiosarcoma_cutaneous_epithelioid = "Malignant:Malignant soft tissue proliferations - Vascular:Angiosarcoma cutaneous:Angiosarcoma cutaneous, Epithelioid"
    malignant_malignant_soft_tissue_proliferations_vascular_angiosarcoma_cutaneous_angiosarcoma_cutaneous_face_and_scalp_of_elderly_patients = "Malignant:Malignant soft tissue proliferations - Vascular:Angiosarcoma cutaneous:Angiosarcoma cutaneous, Face and scalp of elderly patients"
    malignant_malignant_soft_tissue_proliferations_vascular_angiosarcoma_cutaneous_angiosarcoma_cutaneous_post_irradiation = "Malignant:Malignant soft tissue proliferations - Vascular:Angiosarcoma cutaneous:Angiosarcoma cutaneous, Post-irradiation"
    malignant_malignant_soft_tissue_proliferations_vascular_angiosarcoma_cutaneous_angiosarcoma_cutaneous_with_associated_lymphedema = "Malignant:Malignant soft tissue proliferations - Vascular:Angiosarcoma cutaneous:Angiosarcoma cutaneous, With associated lymphedema"
    malignant_malignant_soft_tissue_proliferations_vascular_hemangioendothelioma_hemangioendothelioma_kaposiform = "Malignant:Malignant soft tissue proliferations - Vascular:Hemangioendothelioma:Hemangioendothelioma, Kaposiform"
    benign_benign_melanocytic_proliferations_nevus_blue_nevus_blue_nevus_cellular = (
        "Benign:Benign melanocytic proliferations:Nevus:Blue nevus:Blue nevus, Cellular"
    )
    benign_benign_melanocytic_proliferations_nevus_blue_nevus_blue_nevus_common = (
        "Benign:Benign melanocytic proliferations:Nevus:Blue nevus:Blue nevus, Common"
    )
    benign_benign_melanocytic_proliferations_nevus_blue_nevus_blue_nevus_epithelioid = (
        "Benign:Benign melanocytic proliferations:Nevus:Blue nevus:Blue nevus, Epithelioid"
    )
    benign_benign_melanocytic_proliferations_nevus_blue_nevus_blue_nevus_plaque_type = (
        "Benign:Benign melanocytic proliferations:Nevus:Blue nevus:Blue nevus, Plaque type"
    )
    benign_benign_melanocytic_proliferations_nevus_blue_nevus_blue_nevus_sclerosing = (
        "Benign:Benign melanocytic proliferations:Nevus:Blue nevus:Blue nevus, Sclerosing"
    )
    benign_benign_melanocytic_proliferations_nevus_nevus_atypical_dysplastic_or_clark_nevus_atypical = "Benign:Benign melanocytic proliferations:Nevus:Nevus, Atypical, Dysplastic, or Clark:Nevus, Atypical"
    benign_benign_melanocytic_proliferations_nevus_nevus_atypical_dysplastic_or_clark_nevus_dysplastic = "Benign:Benign melanocytic proliferations:Nevus:Nevus, Atypical, Dysplastic, or Clark:Nevus, Dysplastic"
    benign_benign_melanocytic_proliferations_nevus_nevus_atypical_dysplastic_or_clark_nevus_clark = "Benign:Benign melanocytic proliferations:Nevus:Nevus, Atypical, Dysplastic, or Clark:Nevus, Clark"
    benign_benign_melanocytic_proliferations_nevus_nevus_congenital_nevus_congenital_by_history = "Benign:Benign melanocytic proliferations:Nevus:Nevus, Congenital:Nevus, Congenital, by history"
    benign_benign_melanocytic_proliferations_nevus_nevus_congenital_nevus_congenital_by_histopathological_pattern = "Benign:Benign melanocytic proliferations:Nevus:Nevus, Congenital:Nevus, Congenital, by histopathological pattern"
    benign_benign_melanocytic_proliferations_nevus_nevus_congenital_nevus_congenital_by_history_and_histopathological_pattern = "Benign:Benign melanocytic proliferations:Nevus:Nevus, Congenital:Nevus, Congenital, by history and histopathological pattern"
    benign_mast_cell_proliferations_mastocytosis_mastocytosis_diffuse_or_multifocal_telangiectasia_macularis_eruptiva_perstans = "Benign:Mast cell proliferations:Mastocytosis:Mastocytosis, Diffuse or multifocal:Telangiectasia macularis eruptiva perstans"
    benign_mast_cell_proliferations_mastocytosis_mastocytosis_diffuse_or_multifocal_mastocytosis_diffuse_cutaenous = "Benign:Mast cell proliferations:Mastocytosis:Mastocytosis, Diffuse or multifocal:Mastocytosis, Diffuse cutaenous"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_acantholytic_squamous_cell_carcinoma_invasive_acantholytic_well_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Acantholytic:Squamous cell carcinoma, Invasive, Acantholytic, well differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_acantholytic_squamous_cell_carcinoma_invasive_acantholytic_moderately_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Acantholytic:Squamous cell carcinoma, Invasive, Acantholytic, moderately differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_acantholytic_squamous_cell_carcinoma_invasive_acantholytic_poorly_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Acantholytic:Squamous cell carcinoma, Invasive, Acantholytic, poorly differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_adeno_squamous_squamous_cell_carcinoma_invasive_adeno_squamous_well_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Adeno-squamous:Squamous cell carcinoma, Invasive, Adeno-squamous, well differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_adeno_squamous_squamous_cell_carcinoma_invasive_adeno_squamous_moderately_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Adeno-squamous:Squamous cell carcinoma, Invasive, Adeno-squamous, moderately differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_adeno_squamous_squamous_cell_carcinoma_invasive_adeno_squamous_poorly_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Adeno-squamous:Squamous cell carcinoma, Invasive, Adeno-squamous, poorly differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_clear_cell_squamous_cell_carcinoma_invasive_clear_cell_well_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Clear cell:Squamous cell carcinoma, Invasive, Clear cell, well differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_clear_cell_squamous_cell_carcinoma_invasive_clear_cell_moderately_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Clear cell:Squamous cell carcinoma, Invasive, Clear cell, moderately differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_clear_cell_squamous_cell_carcinoma_invasive_clear_cell_poorly_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Clear cell:Squamous cell carcinoma, Invasive, Clear cell, poorly differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_keratoacanthoma_type_squamous_cell_carcinoma_invasive_keratoacanthoma_type_well_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Keratoacanthoma-type:Squamous cell carcinoma, Invasive, Keratoacanthoma-type, well differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_keratoacanthoma_type_squamous_cell_carcinoma_invasive_keratoacanthoma_type_moderately_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Keratoacanthoma-type:Squamous cell carcinoma, Invasive, Keratoacanthoma-type, moderately differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_keratoacanthoma_type_squamous_cell_carcinoma_invasive_keratoacanthoma_type_poorly_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Keratoacanthoma-type:Squamous cell carcinoma, Invasive, Keratoacanthoma-type, poorly differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_sarcomatoid_squamous_cell_carcinoma_invasive_sarcomatoid_well_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Sarcomatoid:Squamous cell carcinoma, Invasive, Sarcomatoid, well differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_sarcomatoid_squamous_cell_carcinoma_invasive_sarcomatoid_moderately_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Sarcomatoid:Squamous cell carcinoma, Invasive, Sarcomatoid, moderately differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_sarcomatoid_squamous_cell_carcinoma_invasive_sarcomatoid_poorly_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Sarcomatoid:Squamous cell carcinoma, Invasive, Sarcomatoid, poorly differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_spindle_cell_squamous_cell_carcinoma_invasive_spindle_cell_well_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Spindle cell:Squamous cell carcinoma, Invasive, Spindle cell, well differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_spindle_cell_squamous_cell_carcinoma_invasive_spindle_cell_moderately_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Spindle cell:Squamous cell carcinoma, Invasive, Spindle cell, moderately differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_spindle_cell_squamous_cell_carcinoma_invasive_spindle_cell_poorly_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Spindle cell:Squamous cell carcinoma, Invasive, Spindle cell, poorly differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_verrucous_squamous_cell_carcinoma_invasive_verrucous_well_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Verrucous:Squamous cell carcinoma, Invasive, Verrucous, well differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_verrucous_squamous_cell_carcinoma_invasive_verrucous_moderately_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Verrucous:Squamous cell carcinoma, Invasive, Verrucous, moderately differentiated"
    malignant_malignant_epidermal_proliferations_squamous_cell_carcinoma_invasive_squamous_cell_carcinoma_invasive_verrucous_squamous_cell_carcinoma_invasive_verrucous_poorly_differentiated = "Malignant:Malignant epidermal proliferations:Squamous cell carcinoma, Invasive:Squamous cell carcinoma, Invasive, Verrucous:Squamous cell carcinoma, Invasive, Verrucous, poorly differentiated"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_blue_nevus_like_melanoma_invasive_resembling_blue_nevus = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Blue nevus-like:Melanoma Invasive, resembling blue nevus"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_blue_nevus_like_melanoma_invasive_originating_from_blue_nevus = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Blue nevus-like:Melanoma Invasive, originating from blue nevus"
    malignant_malignant_melanocytic_proliferations_melanoma_melanoma_invasive_melanoma_invasive_heavily_pigmented_melanoma_invasive_heavily_pigmented_resembling_epithelioid_blue_nevus_or_melanoma_developing_in_animals = "Malignant:Malignant melanocytic proliferations (Melanoma):Melanoma Invasive:Melanoma Invasive, Heavily pigmented:Melanoma Invasive, Heavily pigmented, resembling epithelioid blue nevus or melanoma developing in animals"

    @staticmethod
    @cache
    def levels(value: str) -> list[str | None]:
        levels_list = value.split(":")
        levels_list += [None] * (5 - len(levels_list))
        return levels_list

    @staticmethod
    @cache
    def as_dict(value: str) -> dict[str, str | None]:
        return {
            "diagnosis_1": DiagnosisEnum.levels(value)[0],
            "diagnosis_2": DiagnosisEnum.levels(value)[1],
            "diagnosis_3": DiagnosisEnum.levels(value)[2],
            "diagnosis_4": DiagnosisEnum.levels(value)[3],
            "diagnosis_5": DiagnosisEnum.levels(value)[4],
        }

    @classmethod
    @cache
    def reverse_ordered_hierarchy(cls) -> list[str]:
        return sorted(cls, key=lambda x: x.value.count(":"), reverse=True)

    @classmethod
    @cache
    def accept_terminal_values(cls, value: Any) -> Any:
        """
        Allow the user to specify any terminal value of the hierarchy to obtain the relevant value.

        e.g. "baz" can be used to obtain "foo:bar:baz", or "bar" for "foo:bar".
        """
        if ":" not in value:
            for member in cls.reverse_ordered_hierarchy():
                if value == member.value.split(":")[-1]:
                    return member.value

        return value

    @classmethod
    @cache
    def _melanoma_diagnoses(cls):
        return [
            diagnosis
            for diagnosis in cls
            if diagnosis.startswith(cls.malignant_malignant_melanocytic_proliferations_melanoma)
        ] + [cls.malignant_collision_at_least_one_malignant_proliferation]

    @classmethod
    @cache
    def is_melanoma(cls, value: str) -> bool:
        return value in cls._melanoma_diagnoses()



from decimal import Decimal
from typing import Any

from hypothesis import given
from hypothesis import strategies as st
from pydantic import ValidationError
import pytest

from isic_metadata.diagnosis_hierarchical import DiagnosisEnum
from isic_metadata.fields import AnatomSiteGeneralEnum
from isic_metadata.metadata import MetadataRow, convert_errors


@pytest.mark.parametrize(
    ("field", "str_value", "parsed_value", "dependent_fields"),
    # dependent_fields is a dict of field names to values are there just to satisfy
    # the field validators.
    [
        ("age", "54", 54, {}),
        ("melanocytic", "True", True, {}),
        ("clin_size_long_diam_mm", "4mm", Decimal("4.0"), {}),
        ("mel_thick_mm", ".33mm", Decimal("0.33"), {"diagnosis": "Melanoma Invasive"}),
        ("mel_ulcer", "false", False, {"diagnosis": "Melanoma Invasive"}),
        ("family_hx_mm", "False", False, {}),
        ("personal_hx_mm", "0", False, {}),
        ("acquisition_day", "142", 142, {}),
    ],
)
def test_non_str_types(
    field: str, str_value: str, parsed_value: Any, dependent_fields: dict[str, Any]
):
    as_str = MetadataRow.model_validate({field: str_value, **dependent_fields})
    as_real = MetadataRow.model_validate({field: parsed_value, **dependent_fields})

    assert as_str.model_dump()[field] == as_real.model_dump()[field]


@pytest.mark.parametrize(("emptyish_value"), ["", " ", "\t", None])
def test_empty_fields_are_omitted(emptyish_value: Any):
    metadata = MetadataRow.model_validate({"diagnosis": "Benign", "mel_ulcer": emptyish_value})
    assert metadata.diagnosis == "Benign"
    assert metadata.mel_thick_mm is None


def test_unstructured_fields():
    metadata = MetadataRow.model_validate({"diagnosis": "Benign", "hello": "world"})
    assert metadata.diagnosis == "Benign"
    assert metadata.unstructured["hello"] == "world"


@pytest.mark.parametrize(("melanoma_diagnosis"), DiagnosisEnum._melanoma_diagnoses())
def test_melanoma_fields(melanoma_diagnosis: str):
    with pytest.raises(ValidationError) as excinfo:
        # mel_ulcer can only be set if diagnosis is melanoma
        MetadataRow.model_validate({"diagnosis": "Benign", "mel_ulcer": True})
    assert len(excinfo.value.errors()) == 1
    assert "mel_ulcer is incompatible with diagnosis" in excinfo.value.errors()[0]["msg"]

    # mel_ulcer can only be set if diagnosis is melanoma
    MetadataRow.model_validate({"diagnosis": melanoma_diagnosis, "mel_ulcer": True})


@given(age=st.integers(min_value=0).map(str))
def test_age_ceiling(age: str):
    metadata = MetadataRow.model_validate({"age": age})
    assert metadata.age is not None
    assert metadata.age <= 85


def test_age_special_case():
    assert MetadataRow.model_validate({"age": "85+"}).age == 85


def test_fitzpatrick_skin_type():
    MetadataRow.model_validate({"fitzpatrick_skin_type": "I"})


@pytest.mark.parametrize(
    ("raw", "parsed"),
    [
        ("4.5 mm", Decimal("4.5")),
        ("14.2   mm", Decimal("14.2")),
        ("4.5mm", Decimal("4.5")),
        ("1mm", Decimal("1.0")),
        ("3.25", Decimal("3.25")),
    ],
)
def test_mel_thick_mm(raw: str, parsed: float):
    metadata = MetadataRow.model_validate({"diagnosis": "Melanoma Invasive", "mel_thick_mm": raw})
    assert metadata.mel_thick_mm == parsed


def test_mel_thick_mm_invalid():
    with pytest.raises(ValidationError) as excinfo:
        MetadataRow.model_validate({"mel_thick_mm": "foo"})
    assert len(excinfo.value.errors()) == 1
    assert "Unable to parse value as a number" in convert_errors(excinfo.value)[0]["msg"]


@given(
    clin_size=st.one_of(
        # keep max values bounded so they don't generate larger than representable decimals.
        # disallow infinity to avoid the string 'inf'.
        st.floats(min_value=0, max_value=1_000_000, exclude_min=True, allow_infinity=False),
        st.integers(min_value=1, max_value=1_000),
    ).map(lambda x: f"{x} mm")
)
def test_clin_size_long_diam_mm_always_rounded(clin_size: str):
    metadata = MetadataRow.model_validate({"clin_size_long_diam_mm": clin_size})
    assert isinstance(metadata.clin_size_long_diam_mm, Decimal)
    assert metadata.clin_size_long_diam_mm == round(metadata.clin_size_long_diam_mm, ndigits=1)


def test_clin_size_long_diam_mm_invalid():
    with pytest.raises(ValidationError) as excinfo:
        MetadataRow.model_validate({"clin_size_long_diam_mm": "foo"})
    assert len(excinfo.value.errors()) == 1
    assert "Unable to parse value as a number" in convert_errors(excinfo.value)[0]["msg"]


@pytest.mark.parametrize(
    ("anatom_site_special", "anatom_site_general_values"),
    [
        ("acral NOS", ["upper extremity", "lower extremity", "palms/soles"]),
        ("nail NOS", ["upper extremity", "lower extremity", "palms/soles"]),
        ("fingernail", ["upper extremity", "palms/soles"]),
        ("toenail", ["lower extremity", "palms/soles"]),
        ("acral palms or soles", ["upper extremity", "lower extremity", "palms/soles"]),
        (
            "oral or genital",
            ["head/neck", "oral/genital", "lower extremity", "anterior torso", "posterior torso"],
        ),
    ],
)
def test_anatom_site_special(anatom_site_special: str, anatom_site_general_values: list[str]):
    for anatom_site_general_value in anatom_site_general_values:
        metadata = MetadataRow.model_validate(
            {
                "anatom_site_special": anatom_site_special,
                "anatom_site_general": anatom_site_general_value,
            }
        )
        assert metadata.anatom_site_special == anatom_site_special
        assert metadata.anatom_site_general == anatom_site_general_value

    for invalid_anatom_site_general in AnatomSiteGeneralEnum:
        if invalid_anatom_site_general.value not in anatom_site_general_values:
            with pytest.raises(ValidationError) as excinfo:
                MetadataRow.model_validate(
                    {
                        "anatom_site_special": anatom_site_special,
                        "anatom_site_general": invalid_anatom_site_general.value,
                    }
                )
            assert len(excinfo.value.errors()) == 1
            assert "is incompatible with anatom_site_general" in excinfo.value.errors()[0]["msg"]


def test_anatom_site_special_requires_anatom_site_general():
    with pytest.raises(ValidationError) as excinfo:
        MetadataRow.model_validate({"anatom_site_special": "acral NOS"})
    assert len(excinfo.value.errors()) == 1
    assert "requires setting anatom_site_general" in excinfo.value.errors()[0]["msg"]



from __future__ import annotations

import contextlib
from dataclasses import dataclass
from importlib.metadata import PackageNotFoundError, version
from typing import Any, Literal

with contextlib.suppress(PackageNotFoundError):
    __version__ = version("isic-metadata")


@dataclass()
class SearchConfig:
    key: str
    es_property: dict[str, Any]
    es_facet: dict[str, Any]


@dataclass()
class Field:
    search: SearchConfig | None = None
    label: str | None = None
    type: Literal["acquisition", "clinical"] | None = None


FIELD_REGISTRY: dict[str, Field] = {}

for field in [
    "blurry",
    "hairy",
    "marker_pen",
    "personal_hx_mm",
    "family_hx_mm",
    "melanocytic",
    "mel_ulcer",
    "concomitant_biopsy",
]:
    FIELD_REGISTRY[field] = Field(
        search=SearchConfig(
            key=field,
            es_property={"type": "boolean"},
            es_facet={"terms": {"field": field}},
        )
    )


for field in [
    "sex",
    "diagnosis_confirm_type",
    "image_type",
    "dermoscopic_type",
    "tbp_tile_type",
    "mel_mitotic_index",
    "anatom_site_general",
    "anatom_site_special",
    "color_tint",
    "patient_id",
    "lesion_id",
    "rcm_case_id",
    "image_manipulation",
]:
    FIELD_REGISTRY[field] = Field(
        search=SearchConfig(
            key=field,
            es_property={"type": "keyword"},
            es_facet={"terms": {"field": field}},
        )
    )


FIELD_REGISTRY.update(
    {
        "fitzpatrick_skin_type": Field(
            search=SearchConfig(
                key="fitzpatrick_skin_type",
                es_property={"type": "keyword"},
                es_facet={
                    "terms": {
                        "field": "fitzpatrick_skin_type",
                        "order": {"_key": "asc"},
                    }
                },
            )
        ),
        "clin_size_long_diam_mm": Field(
            search=SearchConfig(
                key="clin_size_long_diam_mm",
                es_property={"type": "float"},
                es_facet={
                    "histogram": {
                        "field": "clin_size_long_diam_mm",
                        "interval": 10,
                        "extended_bounds": {"min": 0, "max": 100},
                    }
                },
            )
        ),
        "acquisition_day": Field(),
        "diagnosis_1": Field(
            search=SearchConfig(
                key="diagnosis_1",
                es_property={"type": "keyword"},
                es_facet={"terms": {"field": "diagnosis_1", "size": 100}},
            )
        ),
        "diagnosis_2": Field(
            search=SearchConfig(
                key="diagnosis_2",
                es_property={"type": "keyword"},
                es_facet={"terms": {"field": "diagnosis_2", "size": 100}},
            )
        ),
        "diagnosis_3": Field(
            search=SearchConfig(
                key="diagnosis_3",
                es_property={"type": "keyword"},
                es_facet={"terms": {"field": "diagnosis_3", "size": 100}},
            )
        ),
        "diagnosis_4": Field(
            search=SearchConfig(
                key="diagnosis_4",
                es_property={"type": "keyword"},
                es_facet={"terms": {"field": "diagnosis_4", "size": 100}},
            )
        ),
        "diagnosis_5": Field(
            search=SearchConfig(
                key="diagnosis_5",
                es_property={"type": "keyword"},
                es_facet={"terms": {"field": "diagnosis_5", "size": 100}},
            )
        ),
        "mel_thick_mm": Field(
            search=SearchConfig(
                key="mel_thick_mm",
                es_property={"type": "float"},
                es_facet={
                    "range": {
                        "field": "mel_thick_mm",
                        "ranges": [
                            {"from": 0.0, "to": 0.5},
                            {"from": 0.5, "to": 1.0},
                            {"from": 1.0, "to": 1.5},
                            {"from": 1.5, "to": 2.0},
                            {"from": 2.0, "to": 2.5},
                            {"from": 2.5, "to": 3.0},
                            {"from": 3.0, "to": 3.5},
                            {"from": 3.5, "to": 4.0},
                            {"from": 4.0, "to": 4.5},
                            {"from": 4.5, "to": 5.0},
                            {"from": 5.0},
                        ],
                    }
                },
            )
        ),
    }
)

for field in FIELD_REGISTRY:  # noqa: PLC0206
    if field in [
        "blurry",
        "color_tint",
        "dermoscopic_type",
        "hairy",
        "image_type",
        "marker_pen",
        "image_manipulation",
    ]:
        FIELD_REGISTRY[field].type = "acquisition"
    else:
        FIELD_REGISTRY[field].type = "clinical"

for field, label in [
    ("anatom_site_general", "Anatomic Site"),
]:
    FIELD_REGISTRY[field].label = label



from isic_metadata.utils import get_unstructured_columns


def test_get_unstructured_columns():
    assert get_unstructured_columns(["age", "foo", "bar"]) == ["bar", "foo"]


def test_get_unstructured_columns_ignore_filename():
    assert get_unstructured_columns(["age", "foo", "bar", "filename"]) == ["bar", "foo"]



from pydantic import ValidationError
import pytest

from isic_metadata.metadata import MetadataBatch, MetadataRow


def test_batch():
    MetadataBatch(
        items=[
            MetadataRow.model_validate({"sex": "male"}),
            MetadataRow.model_validate({"sex": "male"}),
        ]
    )


def test_lesions_belong_to_same_patient():
    with pytest.raises(ValidationError) as excinfo:
        MetadataBatch(
            items=[
                MetadataRow(lesion_id="foo", patient_id="foopatient"),
                MetadataRow(lesion_id="foo", patient_id="barpatient"),
            ]
        )
    assert len(excinfo.value.errors()) == 1
    assert "belong to multiple patients" in excinfo.value.errors()[0]["msg"]


def test_blank_lesions_dont_belong_to_same_patient():
    MetadataBatch(
        items=[
            MetadataRow(lesion_id="", patient_id="foopatient"),
            MetadataRow(lesion_id="", patient_id="barpatient"),
        ]
    )


def test_rcm_case_has_at_most_one_macroscopic_image():
    with pytest.raises(ValidationError) as excinfo:
        MetadataBatch(
            items=[
                MetadataRow(image_type="RCM: macroscopic", rcm_case_id="foo"),
                MetadataRow(image_type="RCM: macroscopic", rcm_case_id="foo"),
            ]
        )
    assert len(excinfo.value.errors()) == 1
    assert "have multiple macroscopic images" in excinfo.value.errors()[0]["msg"]

    MetadataBatch(
        items=[
            MetadataRow(image_type="RCM: macroscopic", rcm_case_id="foo"),
            MetadataRow(image_type="RCM: macroscopic", rcm_case_id="bar"),
        ]
    )


def test_rcm_cases_belong_to_same_lesion():
    with pytest.raises(ValidationError) as excinfo:
        MetadataBatch(
            items=[
                MetadataRow(
                    rcm_case_id="foo", lesion_id="foolesion", _ignore_rcm_model_checks=True
                ),
                MetadataRow(
                    rcm_case_id="foo", lesion_id="barlesion", _ignore_rcm_model_checks=True
                ),
            ]
        )
    assert len(excinfo.value.errors()) == 1
    assert "belong to multiple lesions" in excinfo.value.errors()[0]["msg"]


def test_blank_rcm_cases_dont_belong_to_same_lesion():
    MetadataBatch(
        items=[
            MetadataRow(rcm_case_id="", lesion_id="foolesion"),
            MetadataRow(rcm_case_id="", lesion_id="barlesion"),
        ]
    )



from typing import Any

from pydantic import ValidationError
import pytest

from isic_metadata.metadata import MetadataRow


@pytest.mark.parametrize("diagnosis", [None, "Basal cell carcinoma"])
@pytest.mark.parametrize(
    ("field_name", "field_value"),
    [
        ("mel_mitotic_index", "4/mm^2"),
        ("mel_thick_mm", "4mm"),
        ("mel_ulcer", True),
    ],
)
def test_melanoma_fields_require_melanoma_diagnosis(
    diagnosis: str | None, field_name: str, field_value: Any
):
    with pytest.raises(ValidationError) as excinfo:
        MetadataRow.model_validate({field_name: field_value, "diagnosis": diagnosis})
    assert len(excinfo.value.errors()) == 1
    assert field_name in excinfo.value.errors()[0]["msg"]

    MetadataRow.model_validate({field_name: field_value, "diagnosis": "Melanoma Invasive"})


@pytest.mark.skip("TODO: https://github.com/ImageMarkup/tracker/issues/141")
def test_diagnosis_confirm_type_requires_diagnosis():
    with pytest.raises(ValidationError) as excinfo:
        MetadataRow.model_validate({"diagnosis_confirm_type": "histopathology"})
    assert len(excinfo.value.errors()) == 1
    assert excinfo.value.errors()[0]["loc"][0] == "diagnosis_confirm_type"

    MetadataRow.model_validate(
        {"diagnosis": "Melanoma Invasive", "diagnosis_confirm_type": "histopathology"}
    )


def test_dermoscopic_type_requires_image_type_dermoscopic():
    with pytest.raises(ValidationError) as excinfo:
        MetadataRow.model_validate({"dermoscopic_type": "contact polarized"})
    assert len(excinfo.value.errors()) == 1
    assert "dermoscopic_type requires setting image_type" in excinfo.value.errors()[0]["msg"]

    MetadataRow.model_validate(
        {"dermoscopic_type": "contact polarized", "image_type": "dermoscopic"}
    )


def test_dermoscopic_type_requires_dermoscopic_image_type():
    with pytest.raises(ValidationError) as excinfo:
        MetadataRow.model_validate(
            {"dermoscopic_type": "contact polarized", "image_type": "clinical: overview"}
        )
    assert len(excinfo.value.errors()) == 1
    assert "dermoscopic_type is incompatible with image_type" in excinfo.value.errors()[0]["msg"]


def test_rcm_case_id_requires_rcm_image_type():
    with pytest.raises(ValidationError) as excinfo:
        MetadataRow.model_validate({"rcm_case_id": "12345"})
    assert len(excinfo.value.errors()) == 1
    assert "rcm_case_id requires setting image_type" in excinfo.value.errors()[0]["msg"]

    MetadataRow.model_validate({"rcm_case_id": "12345", "image_type": "RCM: tile"})


def test_rcm_model_checks_disabling():
    with pytest.raises(ValidationError) as excinfo:
        MetadataRow(rcm_case_id="12345")
    assert len(excinfo.value.errors()) == 1
    assert "rcm_case_id requires setting image_type" in excinfo.value.errors()[0]["msg"]

    MetadataRow(rcm_case_id="12345", _ignore_rcm_model_checks=True)


def test_tbp_tile_type_requires_image_type_tbp_tile():
    with pytest.raises(ValidationError) as excinfo:
        MetadataRow.model_validate({"tbp_tile_type": "2D"})
    assert len(excinfo.value.errors()) == 1
    assert "tbp_tile_type requires setting image_type" in excinfo.value.errors()[0]["msg"]

    MetadataRow.model_validate({"tbp_tile_type": "2D", "image_type": "TBP tile: close-up"})
    MetadataRow.model_validate({"tbp_tile_type": "2D", "image_type": "TBP tile: overview"})


def test_tbp_tile_type_requires_tbp_tile_image_type():
    with pytest.raises(ValidationError) as excinfo:
        MetadataRow.model_validate({"tbp_tile_type": "2D", "image_type": "clinical: overview"})
    assert len(excinfo.value.errors()) == 1
    assert "tbp_tile_type is incompatible with image_type" in excinfo.value.errors()[0]["msg"]

    MetadataRow.model_validate({"tbp_tile_type": "2D", "image_type": "TBP tile: close-up"})
    MetadataRow.model_validate({"tbp_tile_type": "2D", "image_type": "TBP tile: overview"})


def test_concomitant_biopsy_requires_histopathology():
    with pytest.raises(ValidationError) as excinfo:
        MetadataRow.model_validate(
            {"concomitant_biopsy": True, "diagnosis_confirm_type": "single image expert consensus"}
        )
    assert len(excinfo.value.errors()) == 1
    assert (
        "concomitant_biopsy requires setting diagnosis_confirm_type"
        in excinfo.value.errors()[0]["msg"]
    )

    MetadataRow.model_validate(
        {"concomitant_biopsy": True, "diagnosis_confirm_type": "histopathology"}
    )
    MetadataRow.model_validate(
        {"concomitant_biopsy": False, "diagnosis_confirm_type": "single image expert consensus"}
    )



from collections.abc import Sequence

from isic_metadata.metadata import MetadataRow


def get_unstructured_columns(column_names: Sequence[str]) -> list[str]:
    reserved_columns = {"filename", "isic_id"}
    return sorted(set(column_names) - set(MetadataRow.model_fields.keys()) - reserved_columns)



from pathlib import Path

from setuptools import find_packages, setup

readme_file = Path(__file__).parent / "README.md"
with readme_file.open() as f:
    long_description = f.read()

setup(
    name="isic-metadata",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache 2.0",
    url="https://github.com/ImageMarkup/isic-metadata",
    project_urls={
        "Bug Reports": "https://github.com/ImageMarkup/isic-metadata/issues",
        "Source": "https://github.com/ImageMarkup/isic-metadata",
    },
    author="Kitware, Inc.",
    author_email="kitware@kitware.com",
    keywords="requests",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python",
    ],
    python_requires=">=3.10",
    install_requires=["pydantic>=2.4"],
    extras_require={
        "test": [
            "coverage[toml]",
            "hypothesis",
            "pytest",
            "pytest-mock",
            "pytest-cov",
        ],
    },
    packages=find_packages(),
    # needed to py.typed
    include_package_data=True,
)
